import json
import re

from django.db import models
from wagtail.models import CollectionMember
from wagtail.search import index

from . import Audio


class Transcript(CollectionMember, index.Indexed, models.Model):
    audio = models.OneToOneField(Audio, on_delete=models.CASCADE, related_name="transcript")
    podlove = models.FileField(
        upload_to="cast_transcript/",
        null=True,
        blank=True,
        verbose_name="Podlove Transcript",
        help_text="The transcript format for the Podlove Web Player",
    )
    vtt = models.FileField(
        upload_to="cast_transcript/",
        null=True,
        blank=True,
        verbose_name="WebVTT Transcript",
        help_text="The WebVTT format for feed / podcatchers",
    )
    dote = models.FileField(
        upload_to="cast_transcript/",
        null=True,
        blank=True,
        verbose_name="DOTe Transcript",
        help_text="The DOTe json format for feed / podcatchers",
    )

    admin_form_fields: tuple[str, ...] = ("audio", "podlove", "vtt", "dote")

    class Meta:
        ordering = ("-id",)

    @property
    def podlove_data(self) -> dict:
        data = {}
        if self.podlove:
            with self.podlove.open("r") as file:
                data = json.load(file)
        return data

    @property
    def dote_data(self) -> dict:
        data = {}
        if self.dote:
            with self.dote.open("r") as file:
                data = json.load(file)
        return data

    @property
    def podcastindex_data(self) -> dict:
        data = self.dote_data
        if not data:
            return data
        return convert_dote_to_podcastindex_transcript(data)


def time_to_seconds(time_str) -> float:
    match = re.match(r"(\d+):(\d+):(\d+),(\d+)", time_str)
    if match:
        hours, minutes, seconds, milliseconds = map(int, match.groups())
        return hours * 3600 + minutes * 60 + seconds + milliseconds / 1000
    raise ValueError(f"Invalid time format: {time_str}")


def convert_segments(segments) -> list[dict]:
    converted = []
    for segment in segments:
        converted.append(
            {
                "startTime": time_to_seconds(segment["startTime"]),
                "endTime": time_to_seconds(segment["endTime"]),
                "speaker": segment["speakerDesignation"],
                "body": segment["text"],
            }
        )
    return converted


def convert_dote_to_podcastindex_transcript(transcript: dict) -> dict:
    return {
        "version": "1.0",
        "segments": convert_segments(transcript["lines"]),
    }



import json
import logging
from typing import Literal, Union

from rest_framework import serializers

from ..models import Audio, Blog, Video

logger = logging.getLogger(__name__)


class VideoSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:video_detail")
    poster = serializers.ImageField(read_only=True, allow_empty_file=True)
    poster_thumbnail = serializers.ImageField(read_only=True, allow_empty_file=True)

    class Meta:
        model = Video
        fields = ("id", "url", "original", "poster", "poster_thumbnail")


class AudioSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:audio_detail")
    podlove = serializers.HyperlinkedIdentityField(view_name="cast:api:audio_podlove_detail")

    class Meta:
        model = Audio
        fields = ("id", "name", "file_formats", "url", "podlove", "mp3")


class AudioPodloveSerializer(serializers.HyperlinkedModelSerializer):
    version = serializers.SerializerMethodField()
    show = serializers.SerializerMethodField()
    audio = serializers.ListField()
    chapters = serializers.ListField()
    duration = serializers.CharField(source="duration_str")
    link = serializers.URLField(source="episode_url")
    transcripts = serializers.SerializerMethodField()

    class Meta:
        model = Audio
        fields = ("version", "show", "title", "subtitle", "audio", "duration", "chapters", "link", "transcripts")

    def get_show(self, _instance: Audio) -> dict:
        post = self.context.get("post")  # Get the Post object from the context
        if post is None:
            return {}
        episode = post.specific
        podcast = post.blog.specific
        metadata = {
            "title": podcast.title,
            "subtitle": podcast.subtitle,
            "summary": podcast.search_description,  # FIXME is this correct?
            "link": podcast.full_url,
        }
        context = self.context.copy()
        if episode.cover_image is not None:
            context["cover_image_url"] = episode.cover_image.file.url
        cover_image_context = episode.get_cover_image_context(context, podcast)
        metadata["poster"] = cover_image_context["cover_image_url"]
        return metadata

    @staticmethod
    def get_version(_instance: Audio) -> int:
        return 5

    @staticmethod
    def get_transcripts(instance: Audio) -> list[dict]:
        if not hasattr(instance, "transcript"):
            return []
        transcript = instance.transcript
        if transcript.podlove is None:
            return []
        # Open the file and load its contents as JSON
        with transcript.podlove.open("r") as file:
            try:
                data = json.load(file)  # assumes the file content is JSON
                try:
                    transcripts = data["transcripts"]
                except KeyError:
                    # maybe DOTe instead of Podlove -> empty list
                    transcripts = []
                return transcripts
            except json.JSONDecodeError:
                return []


class SimpleBlogSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:facet-counts-detail")

    class Meta:
        model = Blog
        fields = ("id", "url")


FacetName = Literal["date_facets", "category_facets", "tag_facets"]
FacetValueName = Literal["slug", "name", "count"]
FacetList = list[dict[FacetValueName, Union[str, int]]]
FacetCounts = dict[FacetName, FacetList]


class FacetCountSerializer(SimpleBlogSerializer):
    facet_counts = serializers.SerializerMethodField()

    class Meta:
        model = Blog
        fields = ("id", "url", "facet_counts")

    def get_facet_counts(self, instance: Blog) -> FacetCounts:
        """
        Facet counts have the following format:
        {
            "facet_name": [
                {
                    "slug": slug,
                    "name": name,
                    "count": count,
                },
            ],
        }
        """
        get_params = self.context["request"].GET.copy()
        filterset = instance.get_filterset(get_params)

        # transform date facets
        date_facets: FacetList = []
        for date, count in filterset.filters["date_facets"].facet_counts.items():
            year_month: str = date.strftime("%Y-%m")
            date_facets.append({"slug": year_month, "name": year_month, "count": count})

        # transform category facets
        category_facets: FacetList = []
        for slug, (name, count) in filterset.filters["category_facets"].facet_counts.items():
            category_facets.append({"slug": slug, "name": name, "count": count})

        # transform tag facets
        tag_facets: FacetList = []
        for slug, (name, count) in filterset.filters["tag_facets"].facet_counts.items():
            tag_facets.append({"slug": slug, "name": name, "count": count})

        return {
            "date_facets": date_facets,
            "category_facets": category_facets,
            "tag_facets": tag_facets,
        }



from django.http import HttpRequest
from django.urls import include, path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from taggit.models import Tag
from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.admin.panels import FieldPanel
from wagtail.permission_policies.collections import CollectionOwnershipPermissionPolicy
from wagtail.rich_text.pages import PageLinkHandler
from wagtail.snippets.models import register_snippet
from wagtail.snippets.views.snippets import SnippetViewSet

from .admin_urls import audio, transcript, video
from .models import Audio, Transcript, Video


@hooks.register("register_admin_urls")
def register_admin_urls() -> list:
    return [
        path("audio/", include((audio, "castaudio"), namespace="castaudio")),
        path("media/", include((video, "castvideo"), namespace="castvideo")),
        path("transcript/", include((transcript, "cast-transcript"), namespace="cast-transcript")),
    ]


class VideoMenuItem(MenuItem):
    def is_shown(self, request: HttpRequest) -> bool:
        permission_policy = CollectionOwnershipPermissionPolicy(Video, auth_model=Video, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_video_menu_item() -> VideoMenuItem:
    return VideoMenuItem(
        _("Video"),
        reverse("castvideo:index"),
        name="video",
        icon_name="desktop",
        order=300,
    )


class AudioMenuItem(MenuItem):
    def is_shown(self, request: HttpRequest) -> bool:
        permission_policy = CollectionOwnershipPermissionPolicy(Audio, auth_model=Audio, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_audio_menu_item() -> AudioMenuItem:
    return AudioMenuItem(
        _("Audio"),
        reverse("castaudio:index"),
        name="audio",
        icon_name="media",
        order=300,
    )


class TranscriptMenuItem(MenuItem):
    def is_shown(self, request: HttpRequest) -> bool:
        permission_policy = CollectionOwnershipPermissionPolicy(
            Transcript, auth_model=Transcript, owner_field_name="user"
        )
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_transcript_menu_item() -> TranscriptMenuItem:
    return TranscriptMenuItem(
        _("Transcript"),
        reverse("cast-transcript:index"),
        name="transcript",
        icon_name="edit",
        order=300,
    )


@hooks.register("insert_editor_js")
def editor_js_audio_and_video_chooser_urls() -> str:
    """
    This hook is used to insert the audio and video chooser urls. This
    simple code stopped working in Wagtail 6.1:
        <script>
            window.chooserUrls = '{0}';
        </script>
    Throwing this error:
        TypeError: Cannot set properties of undefined (setting 'audioChooser')
            at edit/:100:45
    I don't know why, and I don't have time to investigate. So I'm setting
    both urls in one hook and it works again.
    """
    return format_html(
        """
        <script>
            window.chooserUrls = {{}};
            window.chooserUrls.audioChooser = '{0}';
            window.chooserUrls.videoChooser = '{1}';
        </script>
        """,
        reverse("castaudio:chooser"),
        reverse("castvideo:chooser"),
    )


class TagsSnippetViewSet(SnippetViewSet):
    """
    This is a snippet viewset for the taggit Tag model. It is used to
    be able to do crud operations on tags from the admin menu.
    """

    panels = [FieldPanel("name")]  # only show the name field
    model = Tag
    icon = "tag"  # change as required
    add_to_admin_menu = True
    menu_label = "Tags"
    menu_order = 200  # will put in 3rd place (000 being 1st, 100 2nd)
    list_display = ["name", "slug"]
    search_fields = ("name",)


register_snippet(TagsSnippetViewSet)


class PageLinkHandlerWithCache(PageLinkHandler):
    """
    This is a custom PageLinkHandler that has a cache to store urls
    for internal pages. This is useful when you have all the pages
    anyway in a repository, and you don't want to hit the database
    while rendering links.
    """

    cache: dict[int, str] = {}

    @classmethod
    def cache_url(cls, page_id: int, url: str):
        cls.cache[page_id] = url

    @classmethod
    def expand_db_attributes(cls, attrs):
        if (cached_url := cls.cache.get(int(attrs["id"]))) is not None:
            return f'<a href="{cached_url}">'
        return super().expand_db_attributes(attrs)

    @classmethod
    def expand_db_attributes_many(cls, attrs_list: list[dict]) -> list[str]:
        """Required for Wagtail >= 6.1"""
        links, all_cached = [], True
        for attrs in attrs_list:
            if (cached_url := cls.cache.get(int(attrs["id"]))) is not None:
                links.append(f'<a href="{cached_url}">')
            else:
                all_cached = False
                break
        if all_cached:
            return links
        # if not all cached, fallback to the super method
        return super().expand_db_attributes_many(attrs_list)


@hooks.register("register_rich_text_features")
def register_page_link(features):
    features.register_link_type(PageLinkHandlerWithCache)



# Generated by Django 5.0.4 on 2024-05-08 09:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0055_alter_podcast_itunes_artwork"),
        ("wagtailimages", "0025_alter_image_file_alter_rendition_file"),
    ]

    operations = [
        migrations.AddField(
            model_name="post",
            name="cover",
            field=models.ForeignKey(
                blank=True,
                help_text="An optional cover image.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wagtailimages.image",
            ),
        ),
    ]



# Generated by Django 5.0.3 on 2024-03-13 22:09

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0054_alter_blog_template_base_dir_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="podcast",
            name="itunes_artwork",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="podcasts",
                to="cast.itunesartwork",
            ),
        ),
    ]



"""
Django and Wagtail based blogging / podcasting package
"""

__version__ = "0.2.49"

# Make installation easier by exposing required apps and middleware
from .apps import CAST_APPS, CAST_MIDDLEWARE

__all__ = ["CAST_APPS", "CAST_MIDDLEWARE"]



from abc import abstractmethod
from collections.abc import Iterable
from itertools import chain, islice, tee
from typing import TYPE_CHECKING, Any, Protocol, Union

from django.db.models import Model, QuerySet
from django.template.loader import TemplateDoesNotExist, get_template
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import ClassNotFound, get_lexer_by_name
from wagtail.blocks import CharBlock, ChoiceBlock, ListBlock, StructBlock, TextBlock
from wagtail.blocks.list_block import ListValue
from wagtail.images.blocks import ChooserBlock, ImageChooserBlock
from wagtail.images.models import AbstractImage, AbstractRendition, Image, Rendition

from . import appsettings as settings
from .models.repository import AudioById, ImageById, RenditionsForPosts, VideoById
from .renditions import (
    Height,
    ImageForSlot,
    ImageType,
    Rectangle,
    RenditionFilters,
    Width,
)

if TYPE_CHECKING:
    from .models import Audio, Video
    from .widgets import AdminVideoChooser


def previous_and_next(all_items: Iterable) -> Iterable:
    """
    Turn an iterable into an iterable of tuples of the
    previous and next item in the iterable.

    Example:
        >>> list(previous_and_next(range(3)))
        [(None, 0, 1), (0, 1, 2), (1, 2, None)]
    """
    previous_items, items, next_items = tee(all_items, 3)
    previous_items = chain([None], previous_items)
    next_items = chain(islice(next_items, 1, None), [None])
    return zip(previous_items, items, next_items)


def get_srcset_images_for_slots(
    image: AbstractImage,
    image_type: ImageType,
    renditions: dict[str, AbstractRendition],
) -> dict[Rectangle, ImageForSlot]:
    """
    Get the srcset images for the given slots and image formats. This will return a list of ImageInSlot objects.
    """
    images_for_slots = {}
    rendition_filters = RenditionFilters.from_wagtail_image_with_type(image=image, image_type=image_type)
    slots, image_formats = rendition_filters.slots, rendition_filters.image_formats
    rendition_filters.set_filter_to_url_via_wagtail_renditions(renditions)
    for slot in slots:
        try:
            images_for_slots[slot] = rendition_filters.get_image_for_slot(slot)
        except ValueError:
            # no fitting image found for slot -> use original image
            src = {}
            for image_format in image_formats:
                if image_format == rendition_filters.original_format:
                    src[image_format] = image.file.url
                else:
                    rendition = renditions[f"format-{image_format}"]
                    src[image_format] = rendition.url
            srcset = {}
            for image_format in image_formats:
                if image_format == rendition_filters.original_format:
                    srcset[image_format] = f"{image.file.url} {image.width}w"
                else:
                    rendition = renditions[f"format-{image_format}"]
                    srcset[image_format] = f"{rendition.url} {rendition.width}w"
            width = rendition_filters.slot_to_fitting_width[slot]
            images_for_slots[slot] = ImageForSlot(Rectangle(width, slot.height), src, srcset)
    return images_for_slots


class HasImagesAndRenditions(Protocol):
    image_by_id: ImageById
    renditions_for_posts: RenditionsForPosts


class EmptyImageRepository:
    """
    A repository that can be used if no repository was set.
    """

    image_by_id: ImageById = {}
    renditions_for_posts: RenditionsForPosts = {}


class ChooserGetPrepValueMixin:
    """
    A mixin for the ChooserBlock that allows to set the value to an integer.
    Without this, the page detail api endpoint throws an error because it
    expects an instance or None, but already gets an integer.
    See: https://github.com/ephes/django-cast/issues/157
    """

    def get_prep_value(self, value: Model | int | None) -> int | None:
        if isinstance(value, int):
            return value
        return super().get_prep_value(value)  # type: ignore


class GalleryImageChooserBlock(ImageChooserBlock):
    """
    Fix for https://github.com/ephes/django-cast/issues/157
    """

    def get_prep_value(self, value: Model | dict | None) -> int | None:
        if isinstance(value, dict):
            return value["value"]
        return super().get_prep_value(value)


class CastImageChooserBlock(ChooserGetPrepValueMixin, ImageChooserBlock):
    """
    Just add a thumbnail to the image because we then can use the thumbnail
    to get the srcset and sizes attributes in the template.
    """

    @staticmethod
    def get_image_and_renditions(image_id: int, context: dict) -> tuple[Image, dict[str, Rendition]]:
        repository: HasImagesAndRenditions = context["repository"]
        image = repository.image_by_id[image_id]
        image_renditions = repository.renditions_for_posts.get(image.pk, [])
        fetched_renditions = {r.filter_spec: r for r in image_renditions}
        return image, fetched_renditions

    def get_context(self, image_or_pk: int | Image, parent_context: dict) -> dict:
        assert parent_context is not None
        if isinstance(image_or_pk, int):
            image_pk = image_or_pk
        else:
            # In the preview mode, the image is already an Image object.
            # Do not try to remove this, it's really needed.
            image_pk = image_or_pk.pk
        image, fetched_renditions = self.get_image_and_renditions(image_pk, parent_context)
        images_for_slots = get_srcset_images_for_slots(image, "regular", renditions=fetched_renditions)
        [image.regular] = images_for_slots.values()
        return super().get_context(image, parent_context=parent_context)

    def bulk_to_python(self, items):
        """Overwrite this method to avoid database queries."""
        return items

    def extract_references(self, value):
        yield self.model_class, str(value), "", ""


def add_prev_next(images: Iterable[AbstractImage]) -> None:
    """
    For each image in the queryset, add the previous and next image.
    Note: This function is kept for backward compatibility, but prev/next
    attributes are now calculated directly in templates to handle duplicates correctly.
    """
    for previous_image, current_image, next_image in previous_and_next(images):
        current_image.prev = "false" if previous_image is None else f"img-{previous_image.pk}"
        current_image.next = "false" if next_image is None else f"img-{next_image.pk}"


def add_image_thumbnails(images: Iterable[AbstractImage], context: dict) -> None:
    """
    For each image in the queryset, add the thumbnail and modal image data to the image.
    """
    modal_slot, thumbnail_slot = (
        Rectangle(Width(w), Height(h)) for w, h in settings.CAST_GALLERY_IMAGE_SLOT_DIMENSIONS
    )
    repository: HasRenditionsForPosts = context["repository"]
    renditions_for_posts = repository.renditions_for_posts
    for image in images:
        image_renditions = renditions_for_posts.get(image.pk, [])
        fetched_renditions = {r.filter_spec: r for r in image_renditions}
        images_for_slots = get_srcset_images_for_slots(image, "gallery", renditions=fetched_renditions)
        image.modal = images_for_slots[modal_slot]
        image.thumbnail = images_for_slots[thumbnail_slot]


def prepare_context_for_gallery(images: Iterable[AbstractImage], context: dict) -> dict:
    """
    Add the previous and next image and the thumbnail and modal image data to each
    image of the gallery and then the images to the context.
    """
    images_list = list(images)  # Ensure it's a list
    add_prev_next(images_list)
    add_image_thumbnails(images_list, context=context)
    context["image_pks"] = ",".join([str(image.pk) for image in images_list])
    context["images"] = images_list
    return context


def get_block_template(default_template_name: str, context: dict | None, *, file_name: str) -> str:
    """
    Get the template name for the block from the default template name, the context and the
    local file name. This is needed to be able to overwrite the default template with a
    theme specific template controlled by "template_base_dir" in the context.
    """
    if context is None:
        return default_template_name

    template_base_dir = context.get("template_base_dir")
    if template_base_dir is None:
        return default_template_name

    template_from_theme = f"cast/{template_base_dir}/{file_name}"
    try:
        get_template(template_from_theme)
        return template_from_theme
    except TemplateDoesNotExist:
        return default_template_name


class HasRenditionsForPosts(Protocol):
    renditions_for_posts: RenditionsForPosts


class GalleryProxyRepository:
    """
    A repository that can be used if no repository was set.
    """

    renditions_for_posts: RenditionsForPosts = {}


class GalleryBlock(ListBlock):
    class Meta:
        icon = "image"
        label = "Gallery"
        template = "cast/gallery.html"

    def dict_images_to_id_list(self, values: list[dict]) -> list[int]:
        parsed = []
        for item in values:
            parsed.append(item["value"])
        return parsed

    def get_form_state(self, value):
        """
        I don't know why this is needed, but it is needed to make the editor work
        for galleries. Do not try to remove it, or at least test if galleries are
        still working in the editor. FIXME.
        """
        if isinstance(value, (list, ListValue)):
            not_none_values = list(filter(None, value))
            image_ids = self.dict_images_to_id_list(not_none_values)
            return super().get_form_state(image_ids)
        return super().get_form_state(value)

    def get_template(self, images: QuerySet[AbstractImage] | None = None, context: dict | None = None) -> str:
        default_template_name = super().get_template(images, context)
        return get_block_template(default_template_name, context, file_name="gallery.html")

    def get_context(self, images: QuerySet[AbstractImage], parent_context: dict | None = None) -> dict:
        if parent_context is None:
            parent_context = {"repository": GalleryProxyRepository()}
        context = super().get_context(images, parent_context=parent_context)
        return prepare_context_for_gallery(images, context)


class GalleryBlockWithLayout(StructBlock):
    """
    A gallery block with a layout. The layout parameter controls
    which template is used to render the gallery.
    """

    gallery = GalleryBlock(GalleryImageChooserBlock())
    layout = ChoiceBlock(
        choices=[
            ("default", _("Web Component with Modal")),
            ("htmx", _("HTMX based layout")),
        ],
        default="default",
    )

    class Meta:
        icon = "image"
        label = "Gallery with Layout"
        template = "cast/gallery.html"

    def extract_references(self, value):
        # We don't have the gallery id at this point and not overwriting
        # this method will return a concatenated string of all image items
        # which will lead to postgres complaining about the length of the
        # `wagtailcore_referenceindex.to_object_id` field.
        return []

    def get_template(self, value=None, context=None):
        default_template_name = super().get_template(value, context)
        layout = "default"
        if value is not None:
            if (layout_from_value := value.get("layout")) is not None:
                layout = layout_from_value
        if layout == "htmx":
            template_file_name = "gallery_htmx.html"
        else:
            template_file_name = "gallery.html"
        return get_block_template(default_template_name, context, file_name=template_file_name)

    @staticmethod
    def _get_images_from_repository(repository: HasImagesAndRenditions, values):
        images = []
        for item in values["gallery"]:
            if isinstance(item, dict) and item.get("type") == "item":
                images.append(repository.image_by_id[item["value"]])
            else:
                # it's an Image object
                images.append(item)
        values["gallery"] = images
        return values

    def bulk_to_python(self, values):
        """Postpone the fetching of the database objects to the get_context method."""
        return values

    @staticmethod
    def bulk_to_python_from_database(values):
        image_ids_or_images = list(filter(None, values["gallery"]))
        if len(image_ids_or_images) == 0:
            return values
        if isinstance(image_ids_or_images[0], Image):
            return values
        image_ids_or_images = [item["value"] for item in image_ids_or_images]
        assert isinstance(image_ids_or_images[0], int)
        # we have to fetch the images from the database
        image_ids = image_ids_or_images
        # Fetch all images in one query but preserve the order
        images_by_id = {img.pk: img for img in Image.objects.filter(pk__in=image_ids)}
        # Reconstruct the list in the original order, allowing duplicates
        values["gallery"] = [images_by_id[pk] for pk in image_ids if pk in images_by_id]
        return values

    def from_repository_to_python(self, repository: HasImagesAndRenditions, values):
        try:
            return self._get_images_from_repository(repository, values)
        except KeyError:
            return self.bulk_to_python_from_database(values)

    def get_context(self, value, parent_context: dict | None = None):
        assert parent_context is not None
        repository = parent_context["repository"]
        value = self.from_repository_to_python(repository, value)
        context = super().get_context(value, parent_context=parent_context)
        return prepare_context_for_gallery(value["gallery"], context)


class RepositoryChooserBlock(ChooserGetPrepValueMixin, ChooserBlock):
    def bulk_to_python(self, values):
        """
        Postpone the fetching of the database objects to the get_context method
        because the repository is not available in the bulk_to_python method.
        """
        return values

    def extract_references(self, value):
        yield self.model_class, str(value), "", ""

    @abstractmethod
    def from_repository_to_python(self, repository: Any, value: int) -> Model:
        raise NotImplementedError

    def get_context(self, value, parent_context=None):
        repository = parent_context["repository"]
        value = self.from_repository_to_python(repository, value)
        return super().get_context(value, parent_context=parent_context)


class HasVideos(Protocol):
    video_by_id: VideoById


class VideoChooserBlock(RepositoryChooserBlock):
    @cached_property
    def target_model(self) -> type["Video"]:
        from .models import Video

        return Video

    @cached_property
    def widget(self) -> "AdminVideoChooser":
        from .widgets import AdminVideoChooser

        return AdminVideoChooser()

    def get_form_state(self, value: Union["Video", int] | None) -> dict | None:
        return self.widget.get_value_data(value)

    def from_repository_to_python(self, repository: HasVideos, value: int) -> Model:
        if isinstance(value, self.target_model):
            # this happens in the edit / preview mode
            return value
        try:
            return repository.video_by_id[value]
        except KeyError:
            return super().to_python(value)


class HasAudios(Protocol):
    audio_by_id: AudioById


class AudioChooserBlock(RepositoryChooserBlock):
    def get_template(self, value=None, context=None):
        default_template_name = super().get_template(value, context)
        return get_block_template(default_template_name, context, file_name="audio.html")

    @cached_property
    def target_model(self) -> type["Audio"]:
        from .models import Audio

        return Audio

    @cached_property
    def widget(self):
        from .widgets import AdminAudioChooser

        return AdminAudioChooser()

    def get_form_state(self, value: Union["Audio", int] | None) -> dict | None:
        return self.widget.get_value_data(value)

    def from_repository_to_python(self, repository: HasAudios, value: int) -> Model:
        if isinstance(value, self.target_model):
            # this happens in the edit / preview mode
            return value
        try:
            return repository.audio_by_id[value]
        except KeyError:
            return super().to_python(value)

    def get_prep_value(self, value):
        """
        This is called with an int value when a page with audio is retrieved
        via the API like this: /api/wagtail/pages/67/
        """
        if isinstance(value, int):
            return value
        return super().get_prep_value(value)


class CodeBlock(StructBlock):
    language = CharBlock(help_text="The language of the code block")
    source = TextBlock(rows=8, help_text="The source code of the block")

    def render_basic(self, value: dict | None, context=None) -> str:
        if value is not None:
            try:
                lexer = get_lexer_by_name(value["language"], stripall=True)
            except (ClassNotFound, KeyError):
                lexer = get_lexer_by_name("text", stripall=True)
            highlighted = highlight(value["source"], lexer, HtmlFormatter())
            return mark_safe(highlighted)
        else:
            return ""



from django.db import IntegrityError
from django.http import HttpRequest

from .models import TemplateBaseDirectory

DEFAULT_TEMPLATE_BASE_DIR = "does_not_exist"


def site_template_base_dir(request: HttpRequest) -> dict[str, str]:
    """
    Add the name of the template base directory to the context.
    Add the complete base template path to the context for convenience.
    """
    if hasattr(request, "cast_site_template_base_dir"):
        site_template_base_dir_name = request.cast_site_template_base_dir
    else:
        try:
            site_template_base_dir_name = TemplateBaseDirectory.for_request(request).name
        except (TemplateBaseDirectory.DoesNotExist, IntegrityError):
            # If the site template base directory does not exist, use the default
            # need to catch IntegrityError because of Wagtail5 support
            site_template_base_dir_name = DEFAULT_TEMPLATE_BASE_DIR
    return {
        "cast_site_template_base_dir": site_template_base_dir_name,
        "cast_base_template": f"cast/{site_template_base_dir_name}/base.html",
    }



import logging
from datetime import datetime
from typing import cast

from django.contrib.syndication.views import Feed
from django.db.models import Model, QuerySet
from django.http import Http404, HttpRequest
from django.shortcuts import get_object_or_404
from django.utils.feedgenerator import (
    Atom1Feed,
    Rss201rev2Feed,
    SyndicationFeed,
    rfc2822_date,
)
from django.utils.safestring import SafeText, mark_safe
from wagtail.images.models import Image

from cast import appsettings

from .models import Audio, Blog, Podcast, Post
from .models.repository import FeedRepository
from .views import HtmxHttpRequest

logger = logging.getLogger(__name__)


class RepositoryMixin:
    is_podcast: bool = False
    request: HtmxHttpRequest

    def __init__(self, repository: FeedRepository | None = None) -> None:
        super().__init__()
        self.repository = repository

    def get_repository(self, request: HtmxHttpRequest, blog: Blog) -> FeedRepository:
        if self.repository is not None:
            if not self.repository.used:
                # don't use the same repository twice
                return self.repository  # use predefined repository
        # create new repository
        if appsettings.CAST_REPOSITORY == "default":
            # default repository from cachable data
            cachable_data = FeedRepository.data_for_feed_cachable(
                request=request, blog=blog, is_podcast=self.is_podcast
            )
            return FeedRepository.create_from_cachable_data(data=cachable_data)
        else:
            # create repository from django models
            blog.refresh_from_db()  # FIXME this is stale sometimes
            return FeedRepository.create_from_django_models(
                request=request,
                blog=blog,
                post_queryset=Post.objects.live().descendant_of(blog).order_by("-visible_date"),
            )

    def items(self) -> QuerySet[Post]:
        assert self.repository is not None
        queryset = self.repository.post_queryset
        # mark repository as used - the post_queryset might be empty if used twice
        self.repository.used = True
        return queryset

    def get_feed(self, obj, request) -> SyndicationFeed:
        # If we want to cache the site to avoid one additional db query, we should do it here
        blog = obj
        self.repository = repository = self.get_repository(self.request, blog)
        # now that we have the repository, we can set the template base dir
        # to avoid db queries in context_processors
        self.request.cast_site_template_base_dir = repository.template_base_dir
        feed = super().get_feed(obj, request)  # type: ignore
        feed.repository = repository  # pass repository to feed to be able to access it in PodcastIndexElements
        return feed


class LatestEntriesFeed(RepositoryMixin, Feed):
    object: Blog
    request: HtmxHttpRequest

    def get_object(self, request: HttpRequest, *args, **kwargs) -> Blog:
        self.request = cast(HtmxHttpRequest, request)  # need request for item.serve(request) later on
        slug = kwargs["slug"]
        blog = None
        if self.repository is not None:
            # use predefined repository which is kind of dangerous but makes testing easier.
            # Forgetting to check whether the repository is used led to a bug which resulted in
            # using the wrong blog for the feed :(.
            if not self.repository.used:
                blog = self.repository.blog
        if blog is None:
            blog = get_object_or_404(Blog, slug=slug)
        self.object = blog
        return self.object

    def title(self) -> str:
        return self.object.title

    def description(self) -> str:
        return self.object.description

    def link(self) -> str:
        if self.repository is not None:
            return self.repository.blog_url
        return self.object.get_full_url()

    def item_title(self, post: Model) -> SafeText:
        assert isinstance(post, Post)
        return mark_safe(post.title)

    def item_description(self, post: Model) -> SafeText:
        assert isinstance(post, Post)
        assert self.repository is not None
        repository = self.repository.get_post_detail_repository(post)
        post.description = post.get_description(
            request=self.request, render_detail=True, escape_html=False, repository=repository
        )
        return post.description

    def item_link(self, item) -> SafeText:
        return item.get_full_url()

    def get_context_data(self, **kwargs) -> dict:
        context = super().get_context_data(**kwargs)
        return context


class ITunesElements:
    feed: dict

    def add_artwork(self, podcast: Podcast, handler) -> None:
        if podcast.itunes_artwork is None:
            return

        haqe = handler.addQuickElement
        itunes_artwork_url = cast(Image, podcast.itunes_artwork).original.url
        handler.addQuickElement("itunes:image", attrs={"href": itunes_artwork_url})
        handler.startElement("image", {})
        haqe("url", itunes_artwork_url)
        haqe("title", self.feed["title"])
        handler.endElement("image")

    @staticmethod
    def add_itunes_categories(podcast: Podcast, handler) -> None:
        itunes_categories = podcast.itunes_categories_parsed
        if len(itunes_categories) == 0:
            return
        for category, subcategories in itunes_categories.items():
            handler.startElement("itunes:category", {"text": category})
            for subcategory in subcategories:
                handler.addQuickElement("itunes:category", attrs={"text": subcategory})
            handler.endElement("itunes:category")

    def add_root_elements(self, handler) -> None:
        """Add additional elements to the blog object"""
        super().add_root_elements(handler)  # type: ignore
        haqe = handler.addQuickElement
        blog = self.feed["blog"]

        self.add_artwork(blog, handler)

        haqe("itunes:subtitle", self.feed["subtitle"])
        haqe("itunes:author", blog.author_name)
        handler.startElement("itunes:owner", {})
        haqe("itunes:name", blog.author_name)
        haqe("itunes:email", blog.email)
        handler.endElement("itunes:owner")

        self.add_itunes_categories(blog, handler)

        haqe("itunes:summary", blog.description)
        haqe("itunes:explicit", blog.get_explicit_display())
        try:
            haqe("lastBuildDate", rfc2822_date(blog.last_build_date))
        except IndexError:
            pass
        generator = "Django Web Framework / django-cast"
        haqe("generator", generator)
        haqe("docs", "https://blogs.law.harvard.edu/tech/rss")

    def add_item_elements(self, handler, item) -> None:
        """Add additional elements to the post object"""
        super().add_item_elements(handler, item)  # type: ignore
        haqe = handler.addQuickElement

        post = item["post"]
        haqe("guid", str(post.uuid), attrs={"isPermaLink": "false"})
        # Maybe add license later
        # year = timezone.now().year
        # haqe("copyright", "{0} {1}".format("insert license", year))
        haqe("itunes:author", post.owner.get_full_name())
        haqe("itunes:subtitle", post.title)
        haqe("itunes:summary", post.description)
        haqe("itunes:duration", post.podcast_audio.duration_str)
        haqe("itunes:keywords", post.keywords)
        haqe("itunes:explicit", post.get_explicit_display())
        if post.block:
            haqe("itunes:block", "yes")

    def namespace_attributes(self) -> dict:
        namespace_attributes = {}
        namespace_attributes.update({"xmlns:itunes": "http://www.itunes.com/dtds/podcast-1.0.dtd"})
        return namespace_attributes


class PodcastIndexElements:
    feed: dict
    request: HttpRequest
    repository: FeedRepository

    def add_item_elements(self, handler, item) -> None:
        """Add additional elements to the post object"""
        try:
            super().add_item_elements(handler, item)  # type: ignore
        except AttributeError:
            pass

        haqe = handler.addQuickElement
        episode = item["post"]
        repository = None
        if hasattr(self, "repository"):
            repository = self.repository.get_episode_feed_detail_repository(episode)

        if (vtt_transcript_url := episode.get_vtt_transcript_url(self.request, repository)) is not None:
            haqe("podcast:transcript", attrs={"type": "text/vtt", "url": vtt_transcript_url})
        if (
            podcastindex_transcript_url := episode.get_podcastindex_transcript_url(self.request, repository)
        ) is not None:
            haqe("podcast:transcript", attrs={"type": "application/json", "url": podcastindex_transcript_url})

    def namespace_attributes(self) -> dict:
        namespace_attributes = super().namespace_attributes()  # type: ignore
        namespace_attributes.update({"xmlns:podcast": "https://podcastindex.org/namespace/1.0/"})
        return namespace_attributes


class AtomITunesFeedGenerator(PodcastIndexElements, ITunesElements, Atom1Feed):
    def root_attributes(self) -> dict:
        atom_attrs = super().root_attributes()
        atom_attrs.update(self.namespace_attributes())
        return atom_attrs


class RssITunesFeedGenerator(PodcastIndexElements, ITunesElements, Rss201rev2Feed):
    def rss_attributes(self) -> dict:
        rss_attrs = super().rss_attributes()
        rss_attrs.update(self.namespace_attributes())
        return rss_attrs


class PodcastFeed(RepositoryMixin, Feed):
    """
    A feed of podcasts for iTunes and other compatible podcatchers.
    """

    audio_format: str
    mime_type: str
    object: Podcast
    request: HtmxHttpRequest
    is_podcast: bool = True

    def set_audio_format(self, audio_format: str) -> None:
        format_to_mime = Audio.mime_lookup
        if audio_format not in format_to_mime:
            raise Http404("unknown audio format")
        else:
            self.audio_format = audio_format
            self.mime_type = format_to_mime[audio_format]

    def get_object(self, request, *args, **kwargs) -> Podcast:
        self.set_audio_format(kwargs["audio_format"])

        slug = kwargs["slug"]
        self.object = get_object_or_404(Podcast, slug=slug)
        self.request = request  # need request for item.serve(request) later on
        return self.object

    def link(self) -> str:
        if self.repository is not None:
            return self.repository.blog_url
        return self.object.get_full_url()

    def title(self, _blog: Blog) -> str:
        return self.object.title

    def categories(self, blog: Blog) -> tuple[str]:
        if hasattr(blog, "categories"):
            return (blog.keywords.split(",")[0],)
        else:
            return ("",)

    def itunes_categories(self, blog: Blog) -> list[str]:
        return blog.itunes_categories.split(",")

    def item_title(self, item) -> str:
        return item.title

    def item_description(self, item) -> str:
        repository = None
        if self.repository is not None:
            repository = self.repository.get_post_detail_repository(item)
        item.description = item.get_description(
            request=self.request, render_detail=True, escape_html=False, repository=repository
        )
        return item.description

    def item_link(self, item) -> str:
        return item.get_full_url()

    def item_pubdate(self, item) -> datetime:
        return item.visible_date

    def item_updateddate(self, item: Post) -> datetime:
        return item.last_published_at

    # def item_categories(self, post):
    #    return self.categories(self.blog)

    def item_enclosure_url(self, item: Post) -> str:
        return item.get_enclosure_url(self.audio_format)

    def item_enclosure_length(self, item: Post) -> int:
        return item.get_enclosure_size(self.audio_format)

    def item_enclosure_mime_type(self, _item: Post) -> str:
        return self.mime_type

    def item_keywords(self, item: Post) -> str:
        return item.keywords

    def feed_extra_kwargs(self, obj) -> dict:
        return {"blog": self.object}

    def item_extra_kwargs(self, item) -> dict:
        return {"blog": self.object, "post": item}

    def get_feed(self, obj, request) -> SyndicationFeed:
        feed = super().get_feed(obj, request)
        feed.request = request  # type: ignore
        return feed


class AtomPodcastFeed(PodcastFeed):
    feed_type = AtomITunesFeedGenerator

    def subtitle(self, blog: Blog) -> str:
        return blog.description

    def author_name(self, blog: Blog) -> str:
        return blog.author_name

    def author_email(self, blog) -> str:
        return blog.email

    def link(self) -> str:
        """atom link is still wrong, dunno why FIXME"""
        return self.object.get_full_url()


class RssPodcastFeed(PodcastFeed):
    feed_type = RssITunesFeedGenerator

    def item_guid(self, _post: Post) -> None:
        """ITunesElements can't add isPermaLink attr unless None is returned here."""
        return None

    def description(self, blog: Blog) -> str:
        return blog.description



from django.apps import AppConfig


# Apps required by django-cast
CAST_APPS = [
    # Form styling
    "crispy_forms",
    "crispy_bootstrap4",
    # Filtering and API
    "django_filters",
    "django_htmx",
    "rest_framework",
    # Comments system
    "fluent_comments",  # must be before django_comments
    "threadedcomments",
    "django_comments",
    # Vite support
    "django_vite",
    # Wagtail CMS
    "wagtail.api.v2",
    "wagtail.contrib.forms",
    "wagtail.contrib.redirects",
    "wagtail.contrib.settings",
    "wagtail.embeds",
    "wagtail.sites",
    "wagtail.users",
    "wagtail.snippets",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.search",
    "wagtail.admin",
    "wagtail",
    # Wagtail dependencies
    "modelcluster",
    "taggit",
    # Cast itself
    "cast.apps.CastConfig",
]

# Middleware required by django-cast
CAST_MIDDLEWARE = [
    "django_htmx.middleware.HtmxMiddleware",
    "wagtail.contrib.redirects.middleware.RedirectMiddleware",
]


class CastConfig(AppConfig):
    name: str = "cast"

    def ready(self) -> None:
        from .appsettings import init_cast_settings

        init_cast_settings()



# Generated by Django 5.0.4 on 2024-05-08 20:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0056_add_cover_image_for_post"),
    ]

    operations = [
        migrations.RenameField(
            model_name="post",
            old_name="cover",
            new_name="cover_image",
        ),
        migrations.AddField(
            model_name="post",
            name="cover_alt_text",
            field=models.CharField(blank=True, default="", max_length=255),
        ),
    ]



import logging
import uuid
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any, Optional, Protocol, cast

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.safestring import SafeText
from django.utils.translation import gettext_lazy as _
from django_comments import get_model as get_comment_model
from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from rest_framework.fields import Field
from slugify import slugify
from taggit.models import TaggedItemBase
from wagtail import blocks
from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.admin.panels import FieldPanel, MultiFieldPanel
from wagtail.api import APIField
from wagtail.embeds.blocks import EmbedBlock
from wagtail.fields import StreamField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image, Rendition
from wagtail.models import Page, PageManager
from wagtail.search import index

from cast import appsettings
from cast.blocks import (
    AudioChooserBlock,
    CastImageChooserBlock,
    CodeBlock,
    GalleryBlock,
    GalleryBlockWithLayout,
    VideoChooserBlock,
)
from cast.models import get_or_create_gallery

from ..views import HtmxHttpRequest
from .image_renditions import ImagesWithType, create_missing_renditions_for_posts
from .repository import (
    AudioById,
    EpisodeFeedRepository,
    ImageById,
    LinkTuples,
    PostDetailRepository,
    VideoById,
)
from .theme import TemplateBaseDirectory

if TYPE_CHECKING:
    from .index_pages import Blog, ContextDict, Podcast
    from .transcript import Transcript

logger = logging.getLogger(__name__)


comment_model = get_comment_model()
user_model = get_user_model()


class ContentBlock(blocks.StreamBlock):
    heading: blocks.CharBlock = blocks.CharBlock(classname="full title")
    paragraph: blocks.RichTextBlock = blocks.RichTextBlock()
    code: CodeBlock = CodeBlock(icon="code")
    image: CastImageChooserBlock = CastImageChooserBlock(template="cast/image/image.html")
    gallery: GalleryBlockWithLayout = GalleryBlockWithLayout()
    embed: EmbedBlock = EmbedBlock()
    video: VideoChooserBlock = VideoChooserBlock(template="cast/video/video.html", icon="media")
    audio: AudioChooserBlock = AudioChooserBlock(template="cast/audio/audio.html", icon="media")

    class Meta:
        icon = "form"


TypeToIdSet = dict[str, set[int]]


def sync_media_ids(from_database: TypeToIdSet, from_body: TypeToIdSet) -> tuple[TypeToIdSet, TypeToIdSet]:
    to_add, to_remove = {}, {}
    all_media_types = set(from_database.keys()).union(from_body.keys())
    for media_type in all_media_types:
        in_database_ids = from_database.get(media_type, set())
        in_body_ids = from_body.get(media_type, set())
        ids_to_add = in_body_ids - in_database_ids
        if len(ids_to_add) > 0:
            to_add[media_type] = ids_to_add
        ids_to_remove = in_database_ids - in_body_ids
        if len(ids_to_remove) > 0:
            to_remove[media_type] = ids_to_remove
    return to_add, to_remove


class PostTag(TaggedItemBase):
    content_object = ParentalKey("Post", related_name="tagged_items", on_delete=models.CASCADE)


class HtmlField(Field):
    """
    A serializer field to render the html of a post. It's used in the wagtail api
    to get the rendered html of the overview and detail of a post. An SPA theme
    like `cast_vue` can then use this html instead of having to render the blocks
    itself.
    """

    def __init__(self, *, render_detail: bool = False, **kwargs) -> None:
        self.render_detail = render_detail
        super().__init__(**kwargs)

    def to_representation(self, post: "Post") -> SafeText:
        """
        Pass the request from context to the post's serve method to be able to
        render the post with the correct theme.
        """
        return post.get_description(
            request=self.context["request"], render_detail=self.render_detail, escape_html=False, remove_newlines=False
        )


class HasPostDetails(Protocol):
    post_id: int
    template_base_dir: str
    blog: "Blog"
    root_nav_links: LinkTuples
    comments_are_enabled: bool
    has_audio: bool
    page_url: str
    absolute_page_url: str
    owner_username: str
    blog_url: str
    cover_image_url: str
    cover_alt_text: str
    audio_by_id: AudioById
    video_by_id: VideoById
    image_by_id: ImageById


class Post(Page):
    uuid: models.UUIDField = models.UUIDField(default=uuid.uuid4, editable=False)
    visible_date: models.DateTimeField = models.DateTimeField(
        default=timezone.now,
        help_text=_("The visible date of the post which is used for sorting."),
        db_index=True,
    )
    comments_enabled: models.BooleanField = models.BooleanField(
        default=True,
        help_text=_("Whether comments are enabled for this post."),
    )
    cover_image: models.ForeignKey[Image | None] = models.ForeignKey(
        Image,
        help_text=_("An optional cover image."),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    cover_alt_text: models.CharField = models.CharField(max_length=255, blank=True, default="")

    images: models.ManyToManyField = models.ManyToManyField(Image, blank=True)  # FIXME mypy are you ok?
    videos: models.ManyToManyField = models.ManyToManyField("cast.Video", blank=True)
    galleries: models.ManyToManyField = models.ManyToManyField("cast.Gallery", blank=True)
    audios: models.ManyToManyField = models.ManyToManyField("cast.Audio", blank=True)
    categories = ParentalManyToManyField("cast.PostCategory", blank=True)

    # managers
    tags = ClusterTaggableManager(through=PostTag, blank=True, verbose_name=_("tags"))

    _local_template_name: str | None = None
    _media_lookup: dict[str, dict[int, Any]] | None = None

    # wagtail
    body = StreamField(
        [
            ("overview", ContentBlock()),
            ("detail", ContentBlock()),
        ],
        use_json_field=True,
    )

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]

    api_fields = [
        APIField("uuid"),
        APIField("visible_date"),
        APIField("comments_are_enabled"),
        APIField("cover_image"),
        APIField("cover_alt_text"),
        APIField("body"),
        APIField("html_overview", serializer=HtmlField(source="*", render_detail=False)),
        APIField("html_detail", serializer=HtmlField(source="*", render_detail=True)),
        APIField("comments"),
        APIField("comments_security_data"),
        APIField("podlove_players"),
    ]

    content_panels = Page.content_panels + [
        FieldPanel("visible_date"),
        FieldPanel("categories", widget=forms.CheckboxSelectMultiple),
        FieldPanel("tags"),
        MultiFieldPanel(
            [
                FieldPanel("cover_image"),
                FieldPanel("cover_alt_text"),
            ],
            heading="Cover Image",
            classname="collapsed",
            help_text=_(
                "The cover image for this post. It will be used in the feed, "
                "in the twitter card and maybe on the blog index page."
            ),
        ),
        FieldPanel("body"),
    ]
    parent_page_types = ["cast.Blog", "cast.Podcast"]

    @property
    def media_model_lookup(self) -> dict[str, type[models.Model]]:
        from .audio import Audio
        from .gallery import Gallery
        from .video import Video

        return {
            "image": Image,
            "video": Video,
            "gallery": Gallery,
            "audio": Audio,
        }

    @property
    def blog(self) -> "Blog":
        """
        The get_parent() method returns wagtail parent page, which is not
        necessarily a Blog model, but maybe the root page. If it's a Blog
        it has a .blog attribute containing the model which has all the
        attributes like blog.comments_enabled etc.
        """
        return self.get_parent().blog

    def get_template_base_dir(self, request: HttpRequest) -> str:
        parent = self.get_parent()
        if parent is not None:
            return parent.blog.get_template_base_dir(request)
        else:
            return TemplateBaseDirectory.for_request(request).name

    def get_template(self, request: HttpRequest, *args, local_template_name: str = "post.html", **kwargs) -> str:
        if self._local_template_name is not None:
            local_template_name = self._local_template_name
        template_base_dir = kwargs.get("template_base_dir", None)
        if template_base_dir is None:
            template_base_dir = self.get_template_base_dir(request)
        template = f"cast/{template_base_dir}/{local_template_name}"
        return template

    def __str__(self) -> str:
        return self.title

    def get_slug(self) -> str:
        return slugify(self.title)

    @property
    def media_lookup(self) -> dict[str, dict[int, Any]]:
        if self._media_lookup is not None:
            return self._media_lookup
        try:
            return {
                "image": {i.pk: i for i in self.images.all()},
                "video": {v.pk: v for v in self.videos.all()},
                "gallery": {g.pk: g for g in self.galleries.all()},
                "audio": {a.pk: a for a in self.audios.all()},
            }
        except ValueError:
            # post ist not yet saved
            pass
        return {}

    @property
    def media_attr_lookup(self) -> "ContextDict":
        return {
            "image": self.images,
            "video": self.videos,
            "gallery": self.galleries,
            "audio": self.audios,
        }

    @property
    def audio_in_body(self) -> bool:
        audio_blocks = []
        for block in self.body:
            for content_block in block.value:
                if content_block.block_type == "audio":
                    audio_blocks.append(content_block)
        return len(audio_blocks) > 0

    @property
    def has_audio(self) -> bool:
        audios_count = 0
        try:
            audios_count = self.audios.count()
        except ValueError:
            # will be raised on wagtail preview because page_ptr is not set
            pass
        if self.audio_in_body or audios_count > 0:
            return True
        else:
            if isinstance(self.specific, Episode):
                return self.specific.podcast_audio is not None
            else:
                return False

    def get_comments_are_enabled(self, blog: "Blog") -> bool:
        return appsettings.CAST_COMMENTS_ENABLED and blog.comments_enabled and self.comments_enabled

    @property
    def comments_are_enabled(self) -> bool:
        return self.get_comments_are_enabled(self.blog)

    @property
    def comments_security_data(self) -> dict[str, str | int]:
        from django_comments.forms import CommentSecurityForm

        form = CommentSecurityForm(self)
        return form.generate_security_data()

    @property
    def page_type(self) -> str:
        """
        cannot use wagtail.api.v2.serializers.TypeField.to_representation easily
        """
        return type(self)._meta.app_label + "." + type(self).__name__  # FIXME butt ugly

    def get_url(self, request=None, current_site=None):
        return super().get_url(request=request, current_site=current_site)

    def get_full_url(self, request=None):
        if hasattr(self, "page_url"):
            return self.page_url
        return super().get_full_url(request=request)

    def get_updated_timestamp(self) -> int:
        """Use the last_published_at timestamp if available, otherwise the visible_date."""
        if self.last_published_at is not None:
            return int(self.last_published_at.timestamp())
        else:
            return int(self.visible_date.timestamp())

    @staticmethod
    def get_context_from_repository(context: "ContextDict", repository: HasPostDetails) -> "ContextDict":
        context["template_base_dir"] = repository.template_base_dir
        blog = repository.blog
        context["blog"] = blog
        context["comments_are_enabled"] = repository.comments_are_enabled
        context["root_nav_links"] = repository.root_nav_links
        context["has_audio"] = repository.has_audio
        context["page_url"] = repository.page_url
        context["absolute_page_url"] = repository.absolute_page_url
        context["owner_username"] = repository.owner_username
        context["blog_url"] = repository.blog_url
        context["cover_image_url"] = repository.cover_image_url
        context["cover_alt_text"] = repository.cover_alt_text
        context["audio_items"] = list(repository.audio_by_id.items())
        if context["page"].pk is None:
            context["page"].pk = repository.post_id
        return context

    def get_context(self, request: HttpRequest, **kwargs) -> "ContextDict":
        context = super().get_context(request, **kwargs)
        request = cast(HtmxHttpRequest, request)
        context["repository"] = repository = self.get_repository(request, kwargs)
        context["render_detail"] = kwargs.get("render_detail", False)
        context["render_for_feed"] = kwargs.get("render_for_feed", False)
        context["updated_timestamp"] = self.get_updated_timestamp()
        context = self.get_context_from_repository(context, repository)
        self.owner = user_model(username=context["owner_username"])
        context.update(self.get_cover_image_context(context, blog=context["blog"]))
        if context["render_for_feed"]:
            # use absolute urls for feed
            self.page_url = context["absolute_page_url"]
        else:
            self.page_url = context["page_url"]
        return context

    @property
    def media_ids_from_db(self) -> TypeToIdSet:
        return {k: set(v) for k, v in self.media_lookup.items()}

    def _media_ids_from_body(self, body: StreamField) -> TypeToIdSet:
        from_body: TypeToIdSet = {}
        for content_block in body:
            for block in content_block.value:
                if block.block_type == "gallery":
                    images = block.value.get("gallery", [])
                    image_ids = []
                    for image in images:
                        if isinstance(image, dict):
                            image_ids.append(image["value"])
                        elif isinstance(image, Image):
                            image_ids.append(image.pk)
                        elif isinstance(image, int):
                            image_ids.append(image)
                    media_model = get_or_create_gallery(image_ids)
                else:
                    media_model = block.value
                if block.block_type in self.media_model_lookup:
                    if media_model is not None:
                        if hasattr(media_model, "id"):
                            from_body.setdefault(block.block_type, set()).add(media_model.id)
                        elif isinstance(media_model, int):
                            from_body.setdefault(block.block_type, set()).add(media_model)
                        else:
                            raise ValueError(f"media model {media_model} is not an instance of int or a model")
        return from_body

    @property
    def media_ids_from_body(self) -> TypeToIdSet:
        return self._media_ids_from_body(self.body)

    def sync_media_ids(self) -> None:
        media_attr_lookup = self.media_attr_lookup
        to_add, to_remove = sync_media_ids(self.media_ids_from_db, self.media_ids_from_body)

        # add new ids
        for media_type, ids in to_add.items():
            for media_id in ids:
                media_attr_lookup[media_type].add(media_id)

        # remove obsolete ids
        for media_type, ids in to_remove.items():
            for media_id in ids:
                media_attr_lookup[media_type].remove(media_id)

    # helper methods for image rendition syncing
    @staticmethod
    def get_all_images_from_queryset(posts: Iterable["Post"]) -> ImagesWithType:
        for post in posts:
            yield from post.get_all_images()

    @staticmethod
    def get_all_renditions_from_queryset_flat(posts: Iterable["Post"]) -> Iterator[Rendition]:
        for image_type, image in Post.get_all_images_from_queryset(posts):
            yield from image.renditions.all()

    @staticmethod
    def get_all_renditions_from_queryset(posts: Iterable["Post"]) -> dict[int, list[Rendition]]:
        all_renditions = Post.get_all_renditions_from_queryset_flat(posts)
        renditions_for_posts: dict[int, list[Rendition]] = {}
        for rendition in all_renditions:
            renditions_for_posts.setdefault(rendition.image_id, []).append(rendition)
        return renditions_for_posts

    def get_all_images(self) -> ImagesWithType:
        """
        Use it like this:
        posts_queryset = Post.objects.prefetch_related("images", "galleries__images")[:10]
        for post in posts_queryset:
            for image_type, image in post.get_all_images():
                print(image_type, image)
        """
        try:
            for image in self.images.all():
                yield "regular", image
            for gallery in self.galleries.all():
                for image in gallery.images.all():
                    yield "gallery", image
        except ValueError:
            # will be raised on wagtail preview because page_ptr is not set
            pass

    @property
    def comments(self) -> list[dict[str, int | None | str]]:
        ctype = ContentType.objects.get_for_model(self)
        site_id = getattr(settings, "SITE_ID", None)
        qs = comment_model.objects.filter(
            content_type=ctype,
            object_pk=smart_str(self.pk),
            site__pk=site_id,
        )

        # The is_public and is_removed fields are implementation details of the
        # built-in comment model's spam filtering system, so they might not
        # be present on a custom comment model subclass. If they exist, we
        # should filter on them.
        field_names = [f.name for f in comment_model._meta.fields]
        if "is_public" in field_names:
            qs = qs.filter(is_public=True)
        if getattr(settings, "COMMENTS_HIDE_REMOVED", True) and "is_removed" in field_names:
            qs = qs.filter(is_removed=False)

        result = []
        for comment in qs:
            result.append(
                {
                    "id": comment.id,
                    "parent": comment.parent_id,
                    "user": comment.user_name,
                    "date": comment.submit_date,
                    "comment": comment.comment,
                }
            )
        return result

    @property
    def podlove_players(self) -> list[tuple[str, str]]:
        """
        Get the podlove player data for posts containing audio elements.
        """
        result = []
        for pk, audio in self.media_lookup.get("audio", {}).items():
            element_id = f"#audio_{pk}"
            result.append((element_id, audio.get_podlove_url(self.pk)))
        return result

    @staticmethod
    def get_cover_image_context(context: "ContextDict", blog: Optional["Blog"]) -> dict[str, str]:
        if (
            cover_image_url_from_post := context.get("cover_image_url")
        ) is not None and cover_image_url_from_post != "":
            # if the cover image is set in the context, use it
            return {"cover_image_url": cover_image_url_from_post, "cover_alt_text": context.get("cover_alt_text", "")}
        if blog is not None:
            return blog.get_cover_image_context()

        # no cover image set
        return {"cover_image_url": "", "cover_alt_text": ""}

    def get_description(
        self,
        *,
        request: HttpRequest,
        render_detail: bool = False,
        escape_html: bool = True,
        remove_newlines: bool = True,
        repository: PostDetailRepository | None = None,
    ) -> SafeText:
        """
        Get a description for the feed or twitter player card. Needs to be
        a method because the feed is able to pass the actual request object.
        """
        request = cast(HtmxHttpRequest, request)
        if repository is None:
            repository = self.get_repository(request, {})
        self._local_template_name = "post_body.html"
        description = self.serve(
            request, render_detail=render_detail, repository=repository, render_for_feed=True
        ).rendered_content
        if remove_newlines:
            description = description.replace("\n", "")
        if escape_html:
            description = escape(description)
        return description

    def get_absolute_url(self) -> str:
        """This is needed for django-fluentcomments."""
        return self.full_url

    def get_site(self):
        if hasattr(self, "_repository"):
            return self._repository.site
        return super().get_site()

    def get_repository(self, request: HtmxHttpRequest, kwargs: dict[str, Any]) -> PostDetailRepository:
        repository = kwargs.get("repository")
        if repository is not None:
            return repository
        return PostDetailRepository.create_from_django_models(request=request, post=self)

    def serve(self, request: HtmxHttpRequest, *args, **kwargs):
        kwargs["repository"] = repository = self.get_repository(request, kwargs)
        # set the template_base_dir from the post_data to avoid having self.get_template_base_dir() called
        kwargs["template_base_dir"] = repository.template_base_dir
        return super().serve(request, *args, **kwargs)

    def serve_preview(self, request, mode_name, *args, **kwargs):
        # sync media ids before preview, because otherwise the repository
        # will not have the correct media ids and fail to get the correct
        # renditions and fail with a w1110 not found rendition key error.
        try:
            self.sync_media_ids()  # raises ValueError
            create_missing_renditions_for_posts(iter([self]))  # needed for images src / srcset in preview
        except ValueError:
            # will be raised on wagtail preview because page_ptr is not set
            pass
        return super().serve_preview(request, mode_name)

    def save(self, *args, **kwargs) -> None:
        save_return = super().save(*args, **kwargs)
        self.sync_media_ids()
        create_missing_renditions_for_posts(iter([self]))  # needed for images src / srcset
        return save_return


class CustomEpisodeForm(WagtailAdminPageForm):
    """
    Custom form for Episode to validate the podcast_audio field.

    The reason for this is that the podcast_audio field is not required
    for draft episodes, but it is required for published episodes. So
    we have to check which button was clicked in the admin form.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["action-publish"] = forms.CharField(required=False, widget=forms.HiddenInput())

    def clean(self) -> dict[str, Any]:
        cleaned_data = super().clean()
        if cleaned_data.get("action-publish") and cleaned_data.get("podcast_audio") is None:
            raise forms.ValidationError({"podcast_audio": _("An episode must have an audio file to be published.")})
        return cleaned_data


class Episode(Post):
    """A podcast episode is just a Post with some additional fields."""

    podcast_audio: models.ForeignKey = models.ForeignKey(
        "cast.Audio",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="episodes",
        help_text=_(
            "The audio file for this episode -if this is not set, the episode will not be included in the feed."
        ),
    )
    keywords: models.CharField = models.CharField(
        max_length=255,
        blank=True,
        default="",
        help_text=_(
            """A comma-delimited-list of up to 12 words for iTunes
            searches. Perhaps include misspellings of the title."""
        ),
    )
    EXPLICIT_CHOICES = ((1, _("yes")), (2, _("no")), (3, _("clean")))
    explicit: models.PositiveSmallIntegerField = models.PositiveSmallIntegerField(
        choices=EXPLICIT_CHOICES,
        help_text=_("``Clean`` will put the clean iTunes graphic by it."),
        default=1,
    )
    block: models.BooleanField = models.BooleanField(
        default=False,
        help_text=_(
            "Check to block this episode from iTunes because <br />its "
            "content might cause the entire show to be <br />removed from iTunes."
            ""
        ),
    )

    parent_page_types = ["cast.Podcast"]

    content_panels = Page.content_panels + [
        FieldPanel("visible_date"),
        FieldPanel("podcast_audio"),
        MultiFieldPanel(
            [FieldPanel("categories", widget=forms.CheckboxSelectMultiple)],
            heading="Categories",
            classname="collapsed",
        ),
        FieldPanel("tags"),
        MultiFieldPanel(
            [
                FieldPanel("cover_image"),
                FieldPanel("cover_alt_text"),
            ],
            heading="Cover Image",
            classname="collapsed",
            help_text=_(
                "The cover image for this episode. It will be used in the podcast feed, "
                "in the twitter player card and maybe on the blog index page."
            ),
        ),
        FieldPanel("body"),
        FieldPanel("keywords"),
        FieldPanel("explicit"),
        FieldPanel("block"),
    ]

    objects: PageManager = PageManager()
    aliases_homepage: Any  # FIXME: why is this needed?
    base_form_class = CustomEpisodeForm

    def get_template(self, request: HttpRequest, *args, **kwargs) -> str:
        """
        Use get_template() from the parent class, but pass a local template name.
        """
        return super().get_template(request, *args, local_template_name="episode.html", **kwargs)

    def get_context(self, request, **kwargs) -> "ContextDict":
        context = super().get_context(request, **kwargs)
        context["episode"] = self
        cover_image_context = self.get_cover_image_context(context, self.podcast)
        context.update(cover_image_context)
        if hasattr(request, "build_absolute_uri"):
            blog_slug = context["repository"].blog.slug
            player_url = reverse("cast:twitter-player", kwargs={"episode_slug": self.slug, "blog_slug": blog_slug})
            player_url = request.build_absolute_uri(player_url)
            context["player_url"] = player_url
        return context

    @property
    def podcast(self) -> Optional["Podcast"]:
        """
        The get_parent() method returns wagtail parent page, which is not
        necessarily a Blog model, but maybe the root page. If it's a Blog
        it has a .blog attribute containing the model which has all the
        attributes like blog.comments_enabled etc.
        """
        parent = self.get_parent()
        if parent is not None:
            return parent.specific
        return None

    def get_enclosure_url(self, audio_format: str) -> str:
        return getattr(self.podcast_audio, audio_format).url

    def get_enclosure_size(self, audio_format: str) -> int:
        if self.podcast_audio is None:
            return 0
        from .audio import Audio

        return cast(Audio, self.podcast_audio).get_file_size(audio_format)

    def get_transcript_or_none(self, repository: EpisodeFeedRepository | None) -> Optional["Transcript"]:
        if repository is not None:
            podcast_audio, transcript = repository.podcast_audio, repository.transcript
        else:
            podcast_audio = self.podcast_audio  # type: ignore
            try:
                transcript = podcast_audio.transcript
            except (ObjectDoesNotExist, AttributeError):
                transcript = None
        return transcript

    def get_vtt_transcript_url(self, request: HtmxHttpRequest, repository: EpisodeFeedRepository | None) -> str | None:
        if (transcript := self.get_transcript_or_none(repository)) is not None:
            if transcript.vtt is not None:
                relative_url = reverse("cast:webvtt-transcript", kwargs={"pk": transcript.pk})
                return request.build_absolute_uri(relative_url)
        return None

    def get_podcastindex_transcript_url(
        self, request: HtmxHttpRequest, repository: EpisodeFeedRepository
    ) -> str | None:
        if (transcript := self.get_transcript_or_none(repository)) is not None:
            if transcript.dote is not None:
                relative_url = reverse("cast:podcastindex-transcript-json", kwargs={"pk": transcript.pk})
                return request.build_absolute_uri(relative_url)
        return None


class HomePage(Page):
    body = StreamField(
        [
            ("heading", blocks.CharBlock(classname="full title")),
            ("paragraph", blocks.RichTextBlock()),
            ("image", CastImageChooserBlock(template="cast/image/image.html")),
            ("gallery", GalleryBlock(ImageChooserBlock())),
        ],
        use_json_field=True,
    )
    alias_for_page: models.ForeignKey = models.ForeignKey(
        Page,
        related_name="aliases_homepage",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        verbose_name="Redirect to another page",
        help_text="Make this page an alias for another page, redirecting to it with a non permanent redirect.",
    )

    content_panels = Page.content_panels + [
        FieldPanel("alias_for_page"),
        FieldPanel("body"),
    ]

    def serve(self, request, *args, **kwargs) -> HttpResponse:
        if self.alias_for_page is not None:
            return redirect(cast(Page, self.alias_for_page).url, permanent=False)
        return super().serve(request, *args, **kwargs)



import logging
from collections import OrderedDict
from typing import Any, cast

from django.contrib.auth.mixins import LoginRequiredMixin
from django.db.models import QuerySet
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic import CreateView
from rest_framework import generics, status
from rest_framework.decorators import api_view
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from wagtail.api.v2.router import WagtailAPIRouter
from wagtail.api.v2.views import PagesAPIViewSet
from wagtail.images.api.v2.views import ImagesAPIViewSet

from ..filters import PostFilterset
from ..forms import SelectThemeForm, VideoForm
from ..models import (
    Audio,
    Blog,
    Post,
    SpamFilter,
    Video,
    get_template_base_dir,
    get_template_base_dir_choices,
)
from ..views import HtmxHttpRequest
from ..views.theme import set_template_base_dir
from .serializers import (
    AudioPodloveSerializer,
    AudioSerializer,
    FacetCountSerializer,
    SimpleBlogSerializer,
    VideoSerializer,
)
from .viewmixins import AddRequestUserMixin, FileUploadResponseMixin

logger = logging.getLogger(__name__)


@api_view(["GET"])
def api_root(request: Request) -> Response:
    """
    Show API contents.
    If you add any object types, add them here!
    """
    root_api_urls = (
        # ("images", request.build_absolute_uri(reverse("cast:api:image_list"))),
        # ("galleries", request.build_absolute_uri(reverse("cast:api:gallery_list"))),
        ("videos", request.build_absolute_uri(reverse("cast:api:video_list"))),
        ("audios", request.build_absolute_uri(reverse("cast:api:audio_list"))),
        ("comment_training_data", request.build_absolute_uri(reverse("cast:api:comment-training-data"))),
        ("themes", request.build_absolute_uri(reverse("cast:api:theme-list"))),
        ("pages", request.build_absolute_uri(reverse("cast:api:wagtail:pages:listing"))),
        ("images", request.build_absolute_uri(reverse("cast:api:wagtail:images:listing"))),
        ("facet_counts", request.build_absolute_uri(reverse("cast:api:facet-counts-list"))),
    )
    return Response(OrderedDict(root_api_urls))


class VideoCreateView(LoginRequiredMixin, AddRequestUserMixin, FileUploadResponseMixin, CreateView):  # type: ignore
    model = Video
    form_class = VideoForm
    user_field_name = "user"


class StandardResultsSetPagination(PageNumberPagination):
    page_size = 40
    page_size_query_param = "pageSize"
    max_page_size = 10000


class VideoListView(generics.ListCreateAPIView):
    serializer_class = VideoSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet[Video]:
        user = self.request.user
        qs = Video.objects.all().filter(user=user)
        return qs.order_by("-created")


class VideoDetailView(generics.RetrieveDestroyAPIView):
    queryset = Video.objects.all()
    serializer_class = VideoSerializer
    permission_classes = (IsAuthenticated,)


class AudioListView(generics.ListCreateAPIView):
    serializer_class = AudioSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet[Audio]:
        user = self.request.user
        qs = Audio.objects.all().filter(user=user)
        return qs.order_by("-created")


class AudioDetailView(generics.RetrieveDestroyAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioSerializer
    permission_classes = (IsAuthenticated,)


class AudioPodloveDetailView(generics.RetrieveAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioPodloveSerializer

    def retrieve(self, request: Request, *args, **kwargs) -> Response:
        instance = self.get_object()
        if (episode_id := request.query_params.get("episode_id")) is not None:
            try:
                episode_id = int(episode_id)
                instance.set_episode_id(episode_id)
            except (ValueError, TypeError):
                pass

        # Retrieve post_id from kwargs and add it to context
        if not hasattr(self, "request"):
            # those attributes need to be set before calling get_serializer_context
            self.request = request
            self.format_kwarg = None
        context = self.get_serializer_context()
        post_id = kwargs.get("post_id")
        if post_id:
            post = get_object_or_404(Post, pk=post_id)
            context["post"] = post

        serializer = self.get_serializer(instance, context=context)
        return Response(serializer.data)


class PlayerConfig(generics.RetrieveAPIView):
    def retrieve(self, request: Request, *args, **kwargs) -> Response:
        return Response(
            {
                "activeTab": None,
                "subscribe-button": None,
                "share": {
                    "channels": ["facebook", "twitter", "whats-app", "linkedin", "pinterest", "xing", "mail", "link"],
                    # "outlet": "https://ukw.fm/wp-content/plugins/podlove-web-player/web-player/share.html",
                    "sharePlaytime": True,
                },
                "related-episodes": {"source": "disabled", "value": None},
                "version": 5,
                "theme": {
                    "tokens": {
                        "brand": "#E64415",
                        "brandDark": "#235973",
                        "brandDarkest": "#1A3A4A",
                        "brandLightest": "#E9F1F5",
                        "shadeDark": "#807E7C",
                        "shadeBase": "#807E7C",
                        "contrast": "#000",
                        "alt": "#fff",
                    },
                    "fonts": {
                        "ci": {
                            "name": "ci",
                            "family": [
                                "-apple-system",
                                "BlinkMacSystemFont",
                                "Segoe UI",
                                "Roboto",
                                "Helvetica",
                                "Arial",
                                "sans-serif",
                                "Apple Color Emoji",
                            ],
                            "src": [],
                            "weight": 800,
                        },
                        "regular": {
                            "name": "regular",
                            "family": [
                                "-apple-system",
                                "BlinkMacSystemFont",
                                "Segoe UI",
                                "Roboto",
                                "Helvetica",
                                "Arial",
                                "sans-serif",
                                "Apple Color Emoji",
                            ],
                            "src": [],
                            "weight": 300,
                        },
                        "bold": {
                            "name": "bold",
                            "family": [
                                "-apple-system",
                                "BlinkMacSystemFont",
                                "Segoe UI",
                                "Roboto",
                                "Helvetica",
                                "Arial",
                                "sans-serif",
                                "Apple Color Emoji",
                            ],
                            "src": [],
                            "weight": 700,
                        },
                    },
                },
            }
        )


class FacetCountListView(generics.ListAPIView):
    serializer_class = SimpleBlogSerializer
    queryset = Blog.objects.all().live().order_by("-first_published_at")
    pagination_class = StandardResultsSetPagination


class FacetCountsDetailView(generics.RetrieveAPIView):
    queryset = Blog.objects.all()
    serializer_class = FacetCountSerializer


class CommentTrainingDataView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def get(request, _format: Any = None) -> JsonResponse:
        """
        Return training data for comment classification.
        """
        train = SpamFilter.get_training_data_comments()
        return JsonResponse(train, safe=False)


class ThemeListView(generics.ListAPIView):
    """
    Return a list of available themes. Mark the currently selected theme.
    This is used by the theme switcher for the vue frontend for example.
    """

    def get_queryset(self) -> None:
        return None

    def list(self, request: Request, *args, **kwargs) -> Response:
        choices = get_template_base_dir_choices()
        request = cast(HtmxHttpRequest, request)
        template_base_dir = get_template_base_dir(request, None)
        themes = []
        for slug, name in choices:
            selected = slug == template_base_dir
            themes.append({"slug": slug, "name": name, "selected": selected})
        result = {"items": themes}
        return Response(result)


class UpdateThemeView(APIView):
    """
    Update the selected theme.
    """

    def post(self, request: Request, *args, **kwargs) -> Response:
        if not isinstance(request.data, dict):
            return Response({"error": "Invalid request"}, status=status.HTTP_400_BAD_REQUEST)

        new_theme_slug = request.data.get("theme_slug", None)
        form = SelectThemeForm({"template_base_dir": new_theme_slug})
        if not form.is_valid():
            return Response({"error": "Theme slug is invalid"}, status=status.HTTP_400_BAD_REQUEST)

        request = cast(HtmxHttpRequest, request)
        set_template_base_dir(request, form.cleaned_data["template_base_dir"])

        return Response({"message": "Theme updated successfully"}, status=status.HTTP_200_OK)


class RemoveNullBytesMixin:
    """
    Workaround for query parameters containing null bytes. There
    should be proper input validation in Wagtail APIViewSets, but
    this is a quick fix for now.
    """

    request: HttpRequest

    def cleanup_null_bytes(self):
        for key, value in self.request.GET.items():
            if "\x00" in value:
                mutable_copy = self.request.GET.copy()
                mutable_copy[key] = value.replace("\x00", "")
                self.request.GET = mutable_copy

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        self.cleanup_null_bytes()
        # pycharm gets it, mypy doesn't
        return super().filter_queryset(queryset)  # type: ignore


class FilteredPagesAPIViewSet(RemoveNullBytesMixin, PagesAPIViewSet):
    def get_filtered_queryset(self) -> QuerySet:
        # allow additional query parameters from PostFilterset + use_post_filter flag
        additional_query_params = PostFilterset.Meta.fields + ["use_post_filter"] + ["date_before", "date_after"]
        self.known_query_parameters: set = self.known_query_parameters.union(additional_query_params)
        # remove search parameter from query params because it won't work with PagesAPIViewSet
        # in combination with PostFilterset. But doing full text search on PostFilterset will work.
        original_get_params = self.request.GET.copy()
        get_params = self.request.GET.copy()
        if "search" in get_params:
            del get_params["search"]
        self.request.GET = get_params  # type: ignore
        queryset = super().get_queryset()
        filterset = PostFilterset(data=original_get_params, queryset=queryset)
        return filterset.qs

    def get_queryset(self):
        if self.request.GET.dict().get("use_post_filter", "false") == "true":
            return self.get_filtered_queryset()
        return super().get_queryset()


class CastImagesAPIViewSet(RemoveNullBytesMixin, ImagesAPIViewSet):
    pass


# Wagtail API
wagtail_api_router = WagtailAPIRouter("cast:api:wagtail")
wagtail_api_router.register_endpoint("pages", FilteredPagesAPIViewSet)
wagtail_api_router.register_endpoint("images", CastImagesAPIViewSet)



# Generated by Django 5.1.2 on 2024-11-10 07:26

import django.db.models.deletion
import wagtail.models.media
import wagtail.search.index
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0059_add_subtitle"),
        ("wagtailcore", "0094_alter_page_locale"),
    ]

    operations = [
        migrations.CreateModel(
            name="Transcript",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "podlove",
                    models.FileField(
                        blank=True,
                        help_text="The transcript format for the Podlove Web Player",
                        null=True,
                        upload_to="cast_transcript/",
                        verbose_name="Podlove Transcript",
                    ),
                ),
                (
                    "vtt",
                    models.FileField(
                        blank=True,
                        help_text="The WebVTT format for feed / podcatchers",
                        null=True,
                        upload_to="cast_transcript/",
                        verbose_name="WebVTT Transcript",
                    ),
                ),
                (
                    "dote",
                    models.FileField(
                        blank=True,
                        help_text="The DOTe json format for feed / podcatchers",
                        null=True,
                        upload_to="cast_transcript/",
                        verbose_name="DOTe Transcript",
                    ),
                ),
                (
                    "audio",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, related_name="transcript", to="cast.audio"
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        default=wagtail.models.media.get_root_collection_id,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="wagtailcore.collection",
                        verbose_name="collection",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(wagtail.search.index.Indexed, models.Model),
        ),
    ]



import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, Protocol, TypeAlias, cast

from django.contrib.auth import get_user_model
from django.db import connection
from django.db.models import QuerySet
from django.http import HttpRequest
from wagtail.images.models import Image, Rendition
from wagtail.models import Site

from ..filters import PostFilterset
from ..views import HtmxHttpRequest

if TYPE_CHECKING:
    from cast.blocks import (
        AudioChooserBlock,
        GalleryBlockWithLayout,
        ImageChooserBlock,
        VideoChooserBlock,
    )
    from cast.models import Audio, Blog, Episode, Post, Transcript, Video

PostByID: TypeAlias = dict[int, "Post"]
PageUrlByID: TypeAlias = dict[int, str]
HasAudioByID: TypeAlias = dict[int, bool]
AudiosByPostID: TypeAlias = dict[int, set["Audio"]]
AudioById: TypeAlias = dict[int, "Audio"]
TranscriptByAudioId: TypeAlias = dict[int, "Transcript"]
VideosByPostID: TypeAlias = dict[int, set["Video"]]
VideoById: TypeAlias = dict[int, "Video"]
ImagesByPostID: TypeAlias = dict[int, set["Image"]]
CoverURLByPostID: TypeAlias = dict[int, str]
CoverAltByPostID: TypeAlias = dict[int, str]
ImageById: TypeAlias = dict[int, Image]
RenditionsForPosts: TypeAlias = dict[int, list[Rendition]]
LinkTuples: TypeAlias = list[tuple[str, str]]
RenditionsForPost: TypeAlias = dict[int, list[Rendition]]
SerializedRenditions: TypeAlias = dict[int, list[dict]]
if TYPE_CHECKING:
    CastBlock: TypeAlias = (
        type["AudioChooserBlock"]
        | type["VideoChooserBlock"]
        | type["ImageChooserBlock"]
        | type["GalleryBlockWithLayout"]
    )


def cache_page_url(post_id: int, url: str) -> None:
    from ..wagtail_hooks import PageLinkHandlerWithCache

    PageLinkHandlerWithCache.cache_url(post_id, url)


class QuerysetData:
    """
    This class is a container for the data that is needed to render a list of posts
    and that only depends on the queryset of those posts.
    """

    def __init__(
        self,
        *,
        post_queryset: Any,  # FIXME: Post queryset or list[Post], but does not work
        post_by_id: PostByID,
        audios: AudioById,  # used in blocks
        podcast_audio_by_episode_id: AudioById,  # used in podcast rss feed for podcast:transcript elements
        transcript_by_audio_id: TranscriptByAudioId,  # used in podcast rss feed for podcast:transcript elements
        images: ImageById,
        videos: VideoById,
        audios_by_post_id: AudiosByPostID,
        videos_by_post_id: VideosByPostID,
        images_by_post_id: ImagesByPostID,
        owner_username_by_id: dict[int, str],
        has_audio_by_id: HasAudioByID,
        renditions_for_posts: RenditionsForPost,
        page_url_by_id: PageUrlByID,
        absolute_page_url_by_id: PageUrlByID,
        cover_by_post_id: CoverURLByPostID,
        cover_alt_by_post_id: CoverAltByPostID,
    ):
        self.queryset = post_queryset
        self.post_by_id = post_by_id
        self.audios = audios
        self.images = images
        self.videos = videos
        self.audios_by_post_id = audios_by_post_id
        self.podcast_audio_by_episode_id = podcast_audio_by_episode_id
        self.transcript_by_audio_id = transcript_by_audio_id
        self.videos_by_post_id = videos_by_post_id
        self.images_by_post_id = images_by_post_id
        self.owner_username_by_id = owner_username_by_id
        self.has_audio_by_id = has_audio_by_id
        self.renditions_for_posts = renditions_for_posts
        self.page_url_by_id = page_url_by_id
        self.absolute_page_url_by_id = absolute_page_url_by_id
        self.cover_by_post_id = cover_by_post_id
        self.cover_alt_by_post_id = cover_alt_by_post_id

    @classmethod
    def create_from_post_queryset(
        cls, *, request: HttpRequest, site: Site, queryset: QuerySet["Post"], is_podcast: bool = False
    ) -> "QuerysetData":
        if False:
            queryset = queryset.select_related("owner", "cover_image", "podcast_audio__transcript")
        else:
            queryset = queryset.select_related("owner", "cover_image")
        queryset = queryset.prefetch_related(
            "audios",
            "images",
            "videos",
            "galleries",
            "galleries__images",
            "images__renditions",
            "galleries__images__renditions",
        )
        post_by_id: PostByID = {}
        images, has_audio_by_id, owner_username_by_id, videos, audios = {}, {}, {}, {}, {}
        cover_by_post_id: CoverURLByPostID = {}
        cover_alt_by_post_id: CoverAltByPostID = {}
        audios_by_post_id: AudiosByPostID = {}
        podcast_audio_by_episode_id: AudioById = {}
        transcript_by_audio_id: TranscriptByAudioId = {}
        videos_by_post_id: VideosByPostID = {}
        images_by_post_id: ImagesByPostID = {}
        page_url_by_id: PageUrlByID = {}
        absolute_page_url_by_id: PageUrlByID = {}
        for post in queryset:
            post_by_id[post.pk] = post.specific
            owner_username_by_id[post.pk] = post.owner.username
            has_audio_by_id[post.pk] = post.has_audio
            page_url_by_id[post.pk] = post.get_url(request=request, current_site=site)
            absolute_page_url_by_id[post.pk] = post.full_url
            cover_image_url = ""
            if post.cover_image is not None:
                cover_image_url = cast(Image, post.cover_image).file.url
            cover_by_post_id[post.pk] = cover_image_url
            cover_alt_by_post_id[post.pk] = post.cover_alt_text

            for image_type, image in post.get_all_images():
                images[image.pk] = image
                images_by_post_id.setdefault(post.pk, set()).add(image.pk)
            for video in post.videos.all():
                videos[video.pk] = video
                videos_by_post_id.setdefault(post.pk, set()).add(video.pk)
            for audio in post.audios.all():
                audios[audio.pk] = audio
                audios_by_post_id.setdefault(post.pk, set()).add(audio.pk)
            if hasattr(post, "podcast_audio"):
                podcast_audio_by_episode_id[post.pk] = post.podcast_audio

        from .pages import Post

        return cls(
            post_queryset=queryset,
            post_by_id=post_by_id,
            audios=audios,
            images=images,
            videos=videos,
            audios_by_post_id=audios_by_post_id,
            podcast_audio_by_episode_id=podcast_audio_by_episode_id,
            transcript_by_audio_id=transcript_by_audio_id,
            videos_by_post_id=videos_by_post_id,
            images_by_post_id=images_by_post_id,
            has_audio_by_id=has_audio_by_id,
            renditions_for_posts=Post.get_all_renditions_from_queryset(queryset),
            owner_username_by_id=owner_username_by_id,
            page_url_by_id=page_url_by_id,
            absolute_page_url_by_id=absolute_page_url_by_id,
            cover_by_post_id=cover_by_post_id,
            cover_alt_by_post_id=cover_alt_by_post_id,
        )


class PostDetailRepository:
    """
    This class is a container for the data that is needed to render a post detail page.
    """

    def __init__(
        self,
        *,
        post_id: int,
        template_base_dir: str,
        blog: "Blog",
        root_nav_links: LinkTuples,
        comments_are_enabled: bool,
        has_audio: bool,
        page_url: str,
        absolute_page_url: str,
        owner_username: str,
        blog_url: str,
        cover_image_url: str,
        cover_alt_text: str,
        audio_by_id: AudioById,
        video_by_id: VideoById,
        image_by_id: ImageById,
        renditions_for_posts: RenditionsForPost,
    ):
        self.post_id = post_id
        self.template_base_dir = template_base_dir
        self.blog = blog
        self.root_nav_links = root_nav_links
        self.comments_are_enabled = comments_are_enabled
        self.has_audio = has_audio
        self.page_url = page_url
        self.absolute_page_url = absolute_page_url
        self.owner_username = owner_username
        self.blog_url = blog_url
        self.cover_image_url = cover_image_url
        self.cover_alt_text = cover_alt_text
        self.audio_by_id = audio_by_id
        self.video_by_id = video_by_id
        self.image_by_id = image_by_id
        self.renditions_for_posts = renditions_for_posts

        cache_page_url(post_id, page_url)

    @classmethod
    def create_from_django_models(cls, request: HttpRequest, post: "Post") -> "PostDetailRepository":
        blog = post.blog
        owner_username = "unknown"
        if post.owner is not None:
            owner_username = post.owner.username
        image_by_id = {}  # post.media_lookup.get("image", {}) is not enough because gallery images are missing
        for _, image in post.get_all_images():
            image_by_id[image.pk] = image
        cover_image_url = ""
        if post.cover_image is not None:
            cover_image_url = cast(Image, post.cover_image).file.url
        return cls(
            post_id=post.pk,
            template_base_dir=post.get_template_base_dir(request),
            blog=blog,
            comments_are_enabled=post.get_comments_are_enabled(blog),
            root_nav_links=[(p.get_url(), p.title) for p in blog.get_root().get_children().live()],
            has_audio=post.has_audio,
            page_url=post.get_url(request=request),
            absolute_page_url=post.get_full_url(request=request),
            owner_username=owner_username,
            blog_url=blog.get_url(request=request),
            cover_image_url=cover_image_url,
            cover_alt_text=post.cover_alt_text,
            audio_by_id=post.media_lookup.get("audio", {}),
            video_by_id=post.media_lookup.get("video", {}),
            image_by_id=image_by_id,
            renditions_for_posts=post.get_all_renditions_from_queryset([post]),
        )


class EpisodeFeedRepository:
    """
    This class is a container for the data that is needed to render an episode in the feed.
    """

    def __init__(
        self,
        *,
        podcast_audio: "Audio",
        transcript: Optional["Transcript"],
    ) -> None:
        self.podcast_audio = podcast_audio
        self.transcript = transcript


def audio_to_dict(audio) -> dict:
    data = {
        "id": audio.pk,
        "duration": audio.duration,
        "title": audio.title,
        "subtitle": audio.subtitle,
        "data": audio.data,
        "m4a": audio.m4a.name,
        "mp3": audio.mp3.name,
        "oga": audio.oga.name,
        "opus": audio.opus.name,
    }
    if audio.collection_id is not None:
        data["collection_id"] = audio.collection_id
    else:
        data["collection"] = None
    return data


def transcript_to_dict(transcript) -> dict:
    data = {
        "id": transcript.pk,
        "audio_id": transcript.audio_id,
        "podlove": transcript.podlove.name,
        "vtt": transcript.vtt.name,
        "dote": transcript.dote.name,
    }
    if transcript.collection_id is not None:
        data["collection_id"] = transcript.collection_id
    else:
        data["collection"] = None
    return data


def video_to_dict(video) -> dict:
    data = {
        "id": video.pk,
        "title": video.title,
        "original": video.original.name,
        "poster": video.poster.name,
        "poster_seconds": video.poster_seconds,
    }
    if video.collection_id is not None:
        data["collection_id"] = video.collection_id
    else:
        data["collection"] = None
    return data


def blog_to_dict(blog):
    return {
        "id": blog.pk,
        "pk": blog.pk,
        "author": blog.author,
        "slug": blog.slug,
        "uuid": blog.uuid,
        "email": blog.email,
        "comments_enabled": blog.comments_enabled,
        "noindex": blog.noindex,
        "template_base_dir": blog.template_base_dir,
        "description": blog.description,
    }


def post_to_dict(post):
    return {
        "id": post.pk,
        "pk": post.pk,
        "uuid": post.uuid,
        "slug": post.slug,
        "title": post.title,
        "visible_date": post.visible_date,
        "comments_enabled": post.comments_enabled,
        "body": json.dumps(list(post.body.raw_data)),
    }


def episode_to_dict(post):
    return {
        "id": post.pk,
        "pk": post.pk,
        "uuid": post.uuid,
        "slug": post.slug,
        "title": post.title,
        "visible_date": post.visible_date,
        "comments_enabled": post.comments_enabled,
        "body": json.dumps(list(post.body.raw_data)),
        "podcast_audio": audio_to_dict(post.podcast_audio),
        "keywords": post.keywords,
        "explicit": post.explicit,
        "block": post.block,
    }


def image_to_dict(image):
    data = {
        "pk": image.pk,
        "title": image.title,
        "file": image.file.name,
        "width": image.width,
        "height": image.height,
    }
    if image.collection_id is not None:
        data["collection_id"] = image.collection_id
    else:
        data["collection"] = None
    return data


def rendition_to_dict(rendition):
    return {
        "pk": rendition.pk,
        "filter_spec": rendition.filter_spec,
        "file": rendition.file.name,
        "width": rendition.width,
        "height": rendition.height,
    }


def serialize_renditions(renditions_for_posts: RenditionsForPost) -> SerializedRenditions:
    renditions = {}
    for post_pk, renditions_for_post in renditions_for_posts.items():
        renditions[post_pk] = [rendition_to_dict(rendition) for rendition in renditions_for_post]
    return renditions


def deserialize_renditions(renditions: SerializedRenditions) -> RenditionsForPost:
    return {
        post_pk: [Rendition(**rendition) for rendition in renditions] for post_pk, renditions in renditions.items()
    }


Choice: TypeAlias = tuple[str, str]


class HasChoices(Protocol):
    choices: Iterable[Choice]


def get_facet_choices(fields: dict[str, HasChoices], field_name) -> list[Choice]:
    if field_name in fields:
        return [(k, v) for k, v in fields[field_name].choices if k != ""]
    return []


def add_site_raw(data: dict[str, Any]) -> dict:
    site_statement = """
        select
            id,
            hostname,
            port,
            site_name,
            root_page_id,
            is_default_site
        from
            wagtailcore_site
    """
    with connection.cursor() as cursor:
        cursor.execute(site_statement)
        columns = [col[0] for col in cursor.description]
        row_tuple = cursor.fetchone()
        data["site"] = dict(zip(columns, row_tuple))
    return data


def add_root_nav_links(data: dict[str, Any]) -> dict:
    site = Site(**data["site"])
    root_nav_links = [(p.get_url(), p.title) for p in site.root_page.get_children().live()]
    data["root_nav_links"] = root_nav_links
    return data


def add_queryset_data(data: dict[str, Any], queryset_data: QuerysetData) -> dict:
    # posts
    from .pages import Episode

    post_by_id = {}
    for pk, post in queryset_data.post_by_id.items():
        if isinstance(post, Episode):
            post_by_id[pk] = episode_to_dict(post)
        else:
            post_by_id[pk] = post_to_dict(post)
    data["post_by_id"] = post_by_id

    # audios
    audios = {}
    for pk, audio in queryset_data.audios.items():
        audios[pk] = audio_to_dict(audio)
    data["audios"] = audios

    # transcripts
    transcripts = {}
    for episode_id, audio in queryset_data.podcast_audio_by_episode_id.items():
        if hasattr(audio, "transcript"):
            transcripts[audio.pk] = transcript_to_dict(audio.transcript)

    # videos
    videos = {}
    for pk, video in queryset_data.videos.items():
        videos[pk] = video_to_dict(video)
    data["videos"] = videos

    # images
    images = {}
    for pk, image in queryset_data.images.items():
        images[pk] = image_to_dict(image)
    data["images"] = images

    # renditions
    data["renditions_for_posts"] = serialize_renditions(queryset_data.renditions_for_posts)
    data["posts"] = [post.pk for post in queryset_data.queryset]

    data["images_by_post_id"] = queryset_data.images_by_post_id
    data["videos_by_post_id"] = queryset_data.videos_by_post_id
    data["audios_by_post_id"] = queryset_data.audios_by_post_id
    data["podcast_audio_by_episode_id"] = {
        episode_id: audio_to_dict(audio) for episode_id, audio in queryset_data.podcast_audio_by_episode_id.items()
    }
    data["transcripts"] = transcripts
    data["cover_by_post_id"] = queryset_data.cover_by_post_id
    data["cover_alt_by_post_id"] = queryset_data.cover_alt_by_post_id
    data["has_audio_by_id"] = queryset_data.has_audio_by_id
    data["owner_username_by_id"] = queryset_data.owner_username_by_id
    return data


def data_for_blog_cachable(
    *,
    request: HtmxHttpRequest,
    blog: "Blog",
    is_paginated: bool = True,  # feed is not paginated
    post_queryset: QuerySet["Post"] | None = None,  # queryset is build from filterset / get_params if None
) -> dict:
    """
    Fetch all the data of a blog in a cachable (dict) format.
    """
    data: dict[str, Any] = {"blog": blog_to_dict(blog)}
    data = add_site_raw(data)
    data = add_root_nav_links(data)
    data["template_base_dir"] = blog.get_template_base_dir(request)

    # filters and pagination
    if is_paginated:
        get_params = request.GET.copy()
        filterset = blog.get_filterset(get_params)
        data["filterset"] = {"get_params": get_params.dict()}
        date_facet_choices = [(k, v) for k, v in filterset.form.fields["date_facets"].choices if k != ""]
        data["filterset"]["date_facets_choices"] = date_facet_choices
        data["filterset"]["category_facets_choices"] = get_facet_choices(filterset.form.fields, "category_facets")
        data["filterset"]["tag_facets_choices"] = get_facet_choices(filterset.form.fields, "tag_facets")
        data["pagination_context"] = blog.get_pagination_context(blog.get_published_posts(filterset.qs), get_params)
    # queryset data
    if post_queryset is None:
        post_queryset = data["pagination_context"]["object_list"]
        del data["pagination_context"]["object_list"]  # not cachable
    queryset_data = QuerysetData.create_from_post_queryset(
        request=request, site=Site(**data["site"]), queryset=post_queryset, is_podcast=blog.is_podcast
    )
    data = add_queryset_data(data, queryset_data)

    # page_url by id
    page_url_by_id: PageUrlByID = {}
    absolute_page_url_by_id: PageUrlByID = {}
    for post in queryset_data.queryset:
        page_url_by_id[post.pk] = post.get_url(request=request, current_site=Site(**data["site"]))
        absolute_page_url_by_id[post.pk] = post.full_url
    data["page_url_by_id"] = page_url_by_id
    data["absolute_page_url_by_id"] = absolute_page_url_by_id
    return data


class FeedRepository:
    def __init__(
        self,
        *,  # no positional arguments
        template_base_dir: str = "bootstrap4",
        site: Site,
        blog: "Blog",
        blog_url: str,
        queryset_data: QuerysetData,
        root_nav_links: LinkTuples,
        used: bool = False,
    ):
        self.site = site
        self.blog = blog
        self.blog_url = blog_url
        self.template_base_dir = template_base_dir
        self.root_nav_links = root_nav_links
        self.queryset_data = queryset_data
        self.page_url_by_id = queryset_data.page_url_by_id
        self.absolute_page_url_by_id = queryset_data.absolute_page_url_by_id
        self.renditions_for_posts = queryset_data.renditions_for_posts
        self.images = queryset_data.images
        self.image_by_id = queryset_data.images
        self.post_by_id = queryset_data.post_by_id
        self.owner_username_by_id = queryset_data.owner_username_by_id
        self.has_audio_by_id = queryset_data.has_audio_by_id
        self.videos = queryset_data.videos
        self.audios = queryset_data.audios
        self.audios_by_post_id = queryset_data.audios_by_post_id
        self.used = used
        self.post_queryset = queryset_data.queryset

        for post_id, page_url in self.page_url_by_id.items():
            cache_page_url(post_id, page_url)

    @classmethod
    def create_from_django_models(
        cls,
        *,
        request: HttpRequest,
        blog: "Blog",
        template_base_dir: str = "bootstrap4",
        post_queryset: QuerySet["Post"],
    ) -> "FeedRepository":
        site = Site.find_for_request(request)
        queryset_data = QuerysetData.create_from_post_queryset(request=request, site=site, queryset=post_queryset)
        root_nav_links = [(p.get_url(), p.title) for p in site.root_page.get_children().live()]
        for post in queryset_data.queryset:
            media_lookup: dict[str, dict[int, Audio | Video | Image]] = {}
            for image_pk in queryset_data.images_by_post_id.get(post.pk, []):
                media_lookup.setdefault("image", {}).update({image_pk: queryset_data.images[image_pk]})
            for video_pk in queryset_data.videos_by_post_id.get(post.pk, []):
                media_lookup.setdefault("video", {}).update({video_pk: queryset_data.videos[video_pk]})
            for audio_pk in queryset_data.audios_by_post_id.get(post.pk, []):
                media_lookup.setdefault("audio", {}).update({audio_pk: queryset_data.audios[audio_pk]})
            post._media_lookup = media_lookup

        return cls(
            site=site,
            blog=blog,
            queryset_data=queryset_data,
            template_base_dir=template_base_dir,
            root_nav_links=root_nav_links,
            blog_url=blog.get_url(request=request, current_site=site),
        )

    @staticmethod
    def data_for_feed_cachable(
        *,
        request: HtmxHttpRequest,
        blog: "Blog",
        is_podcast: bool = False,
    ) -> dict:
        blog.refresh_from_db()  # sometimes the blog object is stale / maybe because of serialization? FIXME
        if is_podcast:
            from .pages import Episode

            post_queryset = (
                Episode.objects.live()
                .descendant_of(blog)
                .select_related("podcast_audio__transcript")
                .filter(podcast_audio__isnull=False)
                .order_by("-visible_date")
            )
        else:
            from .pages import Post

            post_queryset = Post.objects.live().descendant_of(blog).order_by("-visible_date")
        data = data_for_blog_cachable(request=request, blog=blog, post_queryset=post_queryset, is_paginated=False)
        data["blog_url"] = blog.get_url(request=request)
        return data

    @classmethod
    def create_from_cachable_data(
        cls,
        *,
        data: dict[str, Any],
    ) -> "FeedRepository":
        """
        This method recreates usable models from the cachable data.
        """
        from wagtail.images.models import Image

        from . import Audio, Blog, Episode, Post, Transcript, Video

        site = Site(**data["site"])
        blog = Blog(**data["blog"])
        template_base_dir = data["template_base_dir"]
        post_by_id = {}
        podcast_fields = ["podcast_audio", "block", "keywords", "explicit"]
        for post_pk, post_data in data["post_by_id"].items():
            is_podcast = any(field in post_data for field in podcast_fields)
            if "podcast_audio" in post_data:
                post_data["podcast_audio"] = Audio(**post_data["podcast_audio"])
            if is_podcast:
                post_by_id[post_pk] = Episode(**post_data)
            else:
                post_by_id[post_pk] = Post(**post_data)
        post_queryset = [post_by_id[post_pk] for post_pk in data["posts"]]
        audios = {audio_pk: Audio(**audio_data) for audio_pk, audio_data in data["audios"].items()}
        images = {image_pk: Image(**image_data) for image_pk, image_data in data["images"].items()}
        videos = {video_pk: Video(**video_data) for video_pk, video_data in data["videos"].items()}
        podcast_audios = {
            episode_pk: Audio(**audio_data)
            for episode_pk, audio_data in data.get("podcast_audio_by_episode_id", {}).items()
        }
        transcripts = {
            audio_pk: Transcript(**transcript_data)
            for audio_pk, transcript_data in data.get("transcripts", {}).items()
        }

        renditions_for_posts = deserialize_renditions(data["renditions_for_posts"])

        user_model = get_user_model()
        for post in post_queryset:
            media_lookup: dict[str, dict[int, Audio | Video | Image]] = {}
            for image_pk in data["images_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("image", {}).update({image_pk: images[image_pk]})
            for video_pk in data["videos_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("video", {}).update({video_pk: videos[video_pk]})
            for audio_pk in data["audios_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("audio", {}).update({audio_pk: audios[audio_pk]})
            post._media_lookup = media_lookup
            post.owner = user_model(username=data["owner_username_by_id"][post.pk])
            post.page_url = data["page_url_by_id"][post.pk]

        queryset_data = QuerysetData(
            post_queryset=post_queryset,
            post_by_id=post_by_id,
            audios=audios,
            images=images,
            videos=videos,
            audios_by_post_id=data["audios_by_post_id"],
            podcast_audio_by_episode_id=podcast_audios,
            transcript_by_audio_id=transcripts,
            videos_by_post_id=data["videos_by_post_id"],
            images_by_post_id=data["images_by_post_id"],
            owner_username_by_id=data["owner_username_by_id"],
            has_audio_by_id=data["has_audio_by_id"],
            renditions_for_posts=renditions_for_posts,
            page_url_by_id=data["page_url_by_id"],
            absolute_page_url_by_id=data["absolute_page_url_by_id"],
            cover_by_post_id=data["cover_by_post_id"],
            cover_alt_by_post_id=data["cover_alt_by_post_id"],
        )
        root_nav_links = data["root_nav_links"]
        return cls(
            **{
                "site": site,
                "blog": blog,
                "blog_url": data["blog_url"],
                "template_base_dir": template_base_dir,
                "queryset_data": queryset_data,
                "root_nav_links": root_nav_links,
            }
        )

    def get_post_detail_repository(self, post: "Post") -> PostDetailRepository:
        post_id = post.id
        blog = self.blog
        return PostDetailRepository(
            post_id=post_id,
            template_base_dir=self.template_base_dir,
            blog=blog,
            root_nav_links=self.root_nav_links,
            comments_are_enabled=post.get_comments_are_enabled(blog),
            has_audio=self.has_audio_by_id[post_id],
            page_url=self.page_url_by_id[post_id],
            absolute_page_url=self.absolute_page_url_by_id[post_id],
            owner_username=self.owner_username_by_id[post_id],
            blog_url=self.blog_url,
            audio_by_id=self.audios,
            video_by_id=self.videos,
            image_by_id=self.images,
            renditions_for_posts=self.renditions_for_posts,
            cover_image_url=self.queryset_data.cover_by_post_id.get(post_id, ""),
            cover_alt_text=self.queryset_data.cover_alt_by_post_id.get(post_id, ""),
        )

    def get_episode_feed_detail_repository(self, episode: "Episode") -> EpisodeFeedRepository:
        episode_id = episode.id
        podcast_audio = self.queryset_data.podcast_audio_by_episode_id[episode_id]
        transcript = self.queryset_data.transcript_by_audio_id.get(podcast_audio.id, None)
        return EpisodeFeedRepository(
            podcast_audio=podcast_audio,
            transcript=transcript,
        )


class BlogIndexRepository:
    def __init__(
        self,
        *,
        template_base_dir: str,
        blog: "Blog",
        filterset: Any,
        queryset_data: QuerysetData,
        pagination_context: dict[str, Any],
        root_nav_links: LinkTuples,
        use_audio_player: bool = False,
    ):
        self.template_base_dir = template_base_dir
        self.blog = blog
        self.filterset = filterset
        self.pagination_context = pagination_context
        self.root_nav_links = root_nav_links
        self.use_audio_player = use_audio_player
        # queryset data
        self.queryset_data = queryset_data
        self.renditions_for_posts = queryset_data.renditions_for_posts
        self.images = queryset_data.images
        self.image_by_id = queryset_data.images
        self.post_by_id = queryset_data.post_by_id
        self.owner_username_by_id = queryset_data.owner_username_by_id
        self.has_audio_by_id = queryset_data.has_audio_by_id
        self.video_by_id = queryset_data.videos
        self.audio_by_id = queryset_data.audios
        self.audios_by_post_id = queryset_data.audios_by_post_id
        self.post_queryset = queryset_data.queryset
        self.page_url_by_id = queryset_data.page_url_by_id
        self.absolute_page_url_by_id = queryset_data.absolute_page_url_by_id

        for post_id, page_url in self.page_url_by_id.items():
            cache_page_url(post_id, page_url)

    @staticmethod
    def data_for_blog_index_cachable(
        *,
        request: HtmxHttpRequest,
        blog: "Blog",
    ) -> dict:
        return data_for_blog_cachable(request=request, blog=blog, is_paginated=True, post_queryset=None)

    @classmethod
    def create_from_cachable_data(
        cls,
        *,
        data: dict[str, Any],
    ) -> "BlogIndexRepository":
        """
        This method recreates usable models from the cachable data.
        """
        from wagtail.images.models import Image

        from . import Audio, Blog, Episode, Post, Video

        # site = Site(**data["site"])
        template_base_dir = data["template_base_dir"]

        post_by_id = {}
        for post_pk, post_data in data["post_by_id"].items():
            if "podcast_audio" in post_data:
                post_data["podcast_audio"] = Audio(**post_data["podcast_audio"])
                post_by_id[post_pk] = Episode(**post_data)
            else:
                post_by_id[post_pk] = Post(**post_data)
        post_queryset = [post_by_id[post_pk] for post_pk in data["posts"]]
        pagination_context = data["pagination_context"]
        pagination_context["object_list"] = post_queryset
        audios = {audio_pk: Audio(**audio_data) for audio_pk, audio_data in data["audios"].items()}
        images = {image_pk: Image(**image_data) for image_pk, image_data in data["images"].items()}
        videos = {video_pk: Video(**video_data) for video_pk, video_data in data["videos"].items()}

        renditions_for_posts = deserialize_renditions(data["renditions_for_posts"])

        user_model = get_user_model()
        use_audio_player = False
        for post in post_queryset:
            media_lookup: dict[str, dict[int, Audio | Video | Image]] = {}
            for image_pk in data["images_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("image", {}).update({image_pk: images[image_pk]})
            for video_pk in data["videos_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("video", {}).update({video_pk: videos[video_pk]})
            for audio_pk in data["audios_by_post_id"].get(post.pk, []):
                media_lookup.setdefault("audio", {}).update({audio_pk: audios[audio_pk]})
            post._media_lookup = media_lookup
            post.owner = user_model(username=data["owner_username_by_id"][post.pk])
            post.page_url = data["page_url_by_id"][post.pk]

            if data["has_audio_by_id"][post.pk]:
                use_audio_player = True

        queryset_data = QuerysetData(
            post_queryset=post_queryset,
            post_by_id=post_by_id,
            audios=audios,
            images=images,
            videos=videos,
            audios_by_post_id=data["audios_by_post_id"],
            podcast_audio_by_episode_id={},  # not needed for blog
            transcript_by_audio_id={},  # not needed for blog
            videos_by_post_id=data["videos_by_post_id"],
            images_by_post_id=data["images_by_post_id"],
            owner_username_by_id=data["owner_username_by_id"],
            has_audio_by_id=data["has_audio_by_id"],
            renditions_for_posts=renditions_for_posts,
            page_url_by_id=data["page_url_by_id"],
            absolute_page_url_by_id=data["absolute_page_url_by_id"],
            cover_by_post_id=data["cover_by_post_id"],
            cover_alt_by_post_id=data["cover_alt_by_post_id"],
        )
        root_nav_links = data["root_nav_links"]

        filterset = PostFilterset(data["filterset"]["get_params"])
        filterset.filters["date_facets"].set_field_choices(data["filterset"]["date_facets_choices"])
        filterset.filters["category_facets"].set_field_choices(data["filterset"]["category_facets_choices"])
        filterset.filters["tag_facets"].set_field_choices(data["filterset"]["tag_facets_choices"])
        delattr(filterset, "_form")

        return cls(
            **{
                # "site": site,
                "blog": Blog(**data["blog"]),
                "template_base_dir": template_base_dir,
                "filterset": filterset,
                "pagination_context": pagination_context,
                "queryset_data": queryset_data,
                "root_nav_links": root_nav_links,
                "use_audio_player": use_audio_player,
            }
        )

    @classmethod
    def create_from_django_models(cls, request: HtmxHttpRequest, blog: "Blog") -> "BlogIndexRepository":
        get_params = request.GET.copy()
        filterset = blog.get_filterset(get_params)
        pagination_context = blog.get_pagination_context(blog.get_published_posts(filterset.qs), get_params)
        use_audio_player = False
        for post in pagination_context["object_list"]:
            post.page_url = post.get_url(request)
            if post.has_audio:
                use_audio_player = True
        template_base_dir = blog.get_template_base_dir(request)
        root_nav_links = []
        site = blog.get_site()
        if site is not None:
            for page in site.root_page.get_children().live():
                root_nav_links.append((page.get_url(request), page.title))
        queryset_data = QuerysetData.create_from_post_queryset(
            request=request, site=site, queryset=pagination_context["object_list"]
        )
        return cls(
            blog=blog,
            filterset=filterset,
            pagination_context=pagination_context,
            template_base_dir=template_base_dir,
            use_audio_player=use_audio_player,
            root_nav_links=root_nav_links,
            queryset_data=queryset_data,
        )
